from datetime import datetime
from numbers import Number
import re
import pandas as pd

try:
    from . import defs
    from . import utils
    from._preproc_utils import MultType
except:
    import defs
    from _preproc_utils import MultType
    import utils


# Age range XX - YY
_p_age_range = re.compile(r"""
    ^                   # Start of string
    (\d+)               # Lower bound of age range. Capture for reuse.
    \s?(-|_|TO)-?\s?          # - or _ between lower and upper bound with optional spaces 
    (\d+)               # Upper bound of age range. Capture for reuse.
    (\s?[-_]\s?\1\s?[-_]\s?\3)?  # Optional mistaken repeat of same pattern. 
    $                   # End of string
    """, re.VERBOSE)

_p_nonlatino = re.compile(r'[\s,]*NON\-?(HISPANIC|LATINO)(\s|$)+')


def convert(converter, col, source_name="", cats=None, std_map=None, delim=None, mult_type=None, item_num=None, 
            no_id="keep", agg_cat=False):
    std_map = {} if std_map is None else std_map
    no_id = no_id.lower()
    if no_id not in ["keep", "null", "error","test"]:
        raise ValueError(f"no_id is {no_id}. It should be 'keep', 'null', or 'error'.")
    
    is_boolean_cols = False
    if len(col.shape)==2: # More than 1 column was passed in
        is_boolean_cols = (col.isnull() | col.isin(['True','False','true','false', True, False,'Yes','No'])).all().all()
        if is_boolean_cols and mult_type!=MultType.SINGLE:
            raise NotImplementedError(f"Multi-column input is not possible for {mult_type} case. "+
                                      "Please submit an issue: https://github.com/openpolicedata/openpolicedata/issues")
        if not is_boolean_cols:
            if col.shape[1]==2:
                # 2 columns are to be merged because 1 or the other is always null
                col_new = col[col.columns[0]].copy()
                is_null = col_new.isnull()
                col_new[is_null] = col[col.columns[1]][is_null]
                col = col_new
            else:
                raise ValueError(f"Unknown multi-column input to convert with columns {col.columns}")
    
    # value_counts handles more data types when getting unique values than .unique()
    counts = col.value_counts(dropna=False)
    vals = counts.index
    if mult_type == MultType.DICT:
        return std_dict(col, std_map, converter, no_id, source_name, cats, agg_cat)
    elif mult_type == MultType.DEMO_COL:
        return std_demo_col(col, vals, std_map, item_num, converter, no_id, source_name, cats, agg_cat)
    elif mult_type == MultType.COUNTS:
        return std_counts(col, vals, std_map, delim, converter, no_id, source_name, cats, agg_cat)
    elif mult_type == MultType.DELIMITED:
        return std_list(col, vals, std_map, delim, converter, no_id, source_name, cats, agg_cat, known_single=True)
    elif mult_type == MultType.WITH_NAME:
        return std_with_names(col, vals, std_map, item_num, converter, no_id, source_name, cats, agg_cat)
    elif mult_type == MultType.WITH_COUNTS:
        return std_with_counts(col, vals, std_map, delim, converter, no_id, source_name, cats, agg_cat)
    else:
        if is_boolean_cols:
            # This currently has only been tested on and needed for Bloomington use of force data
            col_names = col.columns
            vals = []
            for c in col_names:
                words = utils.split_words(c)
                last = ""
                v = []
                is_latino = False
                for w in words:
                    w = w.lower()
                    if last=='non':
                        last = ''
                    elif w=='non':
                        last = w
                    elif w in ['hispanic','latino']:
                        is_latino = True
                    elif w not in ['suspect', 'race', 'gender','perceived']:
                        v.append(w)

                if len(v)==0 and not is_latino:
                    raise ValueError(f"Unable to parse column name {c}")
                v = ' '.join(v)
                if is_latino:
                    v = v +", Latino" if len(v) else 'Latino'
                vals.append(v)

        for x in vals:
            std_map[x] = converter(x, no_id, source_name, cats, agg_cat)
            if isinstance(std_map[x], list):
                std_map[x].sort()
                std_map[x] = ", ".join(std_map[x])
        
        if is_boolean_cols:
            std_map = {c:std_map[v] for c,v in zip(col_names, vals)}
            def convert_bools(x):
                out = set()
                for c in x.index:
                    if x[c] in ['true','True', 'Yes', 'yes'] or x[c]==True:
                        if isinstance(std_map[c],list):
                            out.update(std_map[c])
                        else:
                            out.add(std_map[c])

                out = list(out)
                if len(out)==1:
                    return out[0]
                elif len(out)>1:
                    if agg_cat and defs.get_race_cats()[defs._race_keys.LATINO] in out:
                        return defs.get_race_cats()[defs._race_keys.LATINO]
                    else:
                        return ", ".join(out)
                else:
                    return defs._race_keys.UNSPECIFIED

            return col.apply(convert_bools, axis=1)
        else:
            return col.map(std_map)
    

def convert_off_or_civ(x, no_id, *args, **kwargs):
    orig = x
    if isinstance(x,str):
        x = x.upper()

    if pd.isnull(x) or x in ["MISSING"]:
        return defs._roles.UNSPECIFIED
    elif x in ["OFFICER"]:
        return defs._roles.OFFICER
    elif x in ["SUBJECT","CIVILIAN",'CITIZEN']:
        return defs._roles.SUBJECT
    elif no_id in ["error","test"]:
        raise ValueError(f"Unknown person type {orig}")
    else:
        return orig if no_id=="keep" else ""
    

def _create_age_range_lut(x, no_id, source_name, *args, **kwargs):
    if isinstance(x,str) and "," in x:
        # Look for ages of multiple people
        out = {}
        for k,y in enumerate(x.split(',')):
            try:
                out[k] = _create_age_range_lut(y, "error", source_name)
            except Exception as e:
                break
        else:
            return out

    p_plus = re.compile(r"(\d+)\+\s?-?\s?(\1\+)?",re.IGNORECASE)
    p_plus2 = re.compile(r"(\d+)\s*plus",re.IGNORECASE)
    p_over = re.compile(r"(OVER|>)\s?(\d+)",re.IGNORECASE)
    p_under = re.compile(r"(UNDER|<|BELOW)\s?(\d+)",re.IGNORECASE)
    p_under2 = re.compile(r"(\d+) (AND|&) (UNDER|YOUNGER)",re.IGNORECASE)
    p_above = re.compile(r"(\d+) (AND|&) (ABOVE|OLDER)",re.IGNORECASE)
    p_decade = re.compile(r"^(\d+)s$", re.IGNORECASE)
    p_range = re.compile(r"^(\d+)\s?<(=?)\s?\w+\s?<(=?)\s?(\d+)?$", re.IGNORECASE)

    orig=x
    if type(x)==str:
        x = x.upper().strip()
        x = re.sub(r'[A-Z] \- ','', x)  # Oakland Stops data includes a category label up front such as: 'B - 18-29'. Remove this.
    if isinstance(x,str) and _p_age_range.search(x)!=None:
        return _p_age_range.sub(r"\1-\3", x)
    elif isinstance(x,str) and p_over.search(x)!=None:
        return p_over.sub(r"\2-120", x)
    elif isinstance(x,str) and p_plus.search(x)!=None:
        return p_plus.sub(r"\1-120", x)
    elif isinstance(x,str) and p_plus2.search(x)!=None:
        return p_plus2.sub(r"\1-120", x)
    elif isinstance(x,str) and p_above.search(x)!=None:
        return p_above.sub(r"\1-120", x)
    elif isinstance(x,str) and p_under.search(x)!=None:
        return p_under.sub(r"0-\2", x)
    elif isinstance(x,str) and p_under2.search(x)!=None:
        return p_under2.sub(r"0-\1", x)
    elif isinstance(x,str) and p_decade.search(x)!=None:
        decade = int(p_decade.search(x).group(1))
        return f"{decade}-{decade+9}"
    elif isinstance(x,str) and p_range.search(x)!=None:
        results = p_range.search(x).groups()
        start = int(results[0]) if results[1]=="=" else int(results[0])+1
        stop = int(results[3]) if results[2]=="=" else int(results[3])-1
        return f"{start}-{stop}"
    elif pd.isnull(x) or (isinstance(x,str) and (x.replace(" ","") in ["","NR","-"] or "NO DATA" in x)):
        return ""
    elif x=="ADULT":
        return "18-120"
    elif x=="JUVENILE":
        return "0-17"
    elif isinstance(x,Number) or x.isdigit() or x in ["UNKNOWN","N/A"]:
        return orig
    else:
        # At least 1 case was found where age range was auto-corrected to a month/year like
        # 10-17 to 17-Oct. Check for this before giving up.
        try:
            return datetime.strftime(datetime.strptime(x, "%d-%b"),"%m-%d")
        except:
            if no_id=="keep" or (no_id=="test" and x in ["#VALUE!", "NOT AVAILABLE", "OTHER"]):
                return orig
            elif no_id=="error":
                raise ValueError(f"Unknown val {x} for age range")
            else:
                # no_id == "null"
                return ""
    

def _create_ethnicity_lut(x, no_id, source_name, eth_cats, *args, **kwargs):
    # The below values is used in the Ferndale demographics data. Just use the data from the race column in that
    # case which includes if officer is Hispanic
    ferndale_eth_vals = ['NR', 'FRENCH/GERMAN', 'MEXICAN', 'HUNGARIAN', 'LEBANESE', 'POLISH/SCOTTISH', 'IRISH', 'SYRIAN', 'POLISH']
    orig = x
    if isinstance(x,str):
        # Look for {Full name} {- or =} {Abbreviation_Initial}
        abbrev_full_match = re.search(r"^([\w\s/\.-]+)\s?[-]\s?([\w\s/\.])$",x)
        if abbrev_full_match and any([len(x)==1 for x in abbrev_full_match.groups()]):
            x = [x for x in abbrev_full_match.groups() if len(x)>1][0]

        x = x.upper().replace("-","").replace(" ", "").strip()

    has_unspecified = defs._eth_keys.UNSPECIFIED in eth_cats
    has_nonlat = defs._eth_keys.NONLATINO in eth_cats
    has_latino = defs._eth_keys.LATINO in eth_cats

    if pd.notnull(x):
        # Check if value is part of eth_cats
        # This is particularly necessary for custom eth_cats dicts
        for k,v in eth_cats.items():
            # Key is always be a string
            if k.upper()==str(orig).upper() or k.upper()==str(x) or str(v)==str(orig).upper() or str(v)==str(x):
                return v
        
    if pd.isnull(x):
        if has_unspecified:
            return eth_cats[defs._eth_keys.UNSPECIFIED]
        elif no_id=="error":
            raise ValueError(f"Null value found in ethnicity column: {orig}")
        else:
            # Same result whether no_id is keep or null
            return ""
        
    if has_nonlat and (x in ["N", "NH", "NHIS",'NONE'] or "NOTHISP" in x or "NONHIS" in x or "NONLATINO" in x):
        return eth_cats[defs._eth_keys.NONLATINO]
    if has_latino and (x in ["H","HIS","LAT"] or "HISPANIC" in x or "LATINO" in x):
        return eth_cats[defs._eth_keys.LATINO]
    if defs._eth_keys.UNKNOWN in eth_cats and x in ["U", "UNKNOWN"]:
        return eth_cats[defs._eth_keys.UNKNOWN]
    if defs._eth_keys.UNSPECIFIED in eth_cats and ("NODATA" in x or x in ["MISSING", "", "NOTREPORTED",'NOTAPPICABLE(NONINDIVIDUAL)']):
        return eth_cats[defs._eth_keys.UNSPECIFIED]
    if defs._eth_keys.MIDDLE_EASTERN and x in ["M"] and source_name == "Connecticut":
        return eth_cats[defs._eth_keys.MIDDLE_EASTERN]
    if no_id=="test":
        if "EXEMPT" in x:
            return orig
        elif x in ["W","A"] or x in ferndale_eth_vals:
            return eth_cats[defs._eth_keys.NONLATINO]
    elif no_id=="error":
        raise ValueError(f"Unknown value in ethnicity column: {orig}")
    else:
        return orig if no_id=="keep" else ""


def _create_race_lut(x, no_id, source_name, race_cats=defs.get_race_cats(), agg_cat=False, known_single=False):
    if race_cats is None:
        race_cats = defs.get_race_cats()

    # Create flags to indicate which keys are in race_cats
    has_unspecified = defs._race_keys.UNSPECIFIED in race_cats
    has_aapi = defs._race_keys.AAPI in race_cats
    has_asian = defs._race_keys.ASIAN in race_cats
    has_black = defs._race_keys.BLACK in race_cats
    has_indigenous = defs._race_keys.INDIGENOUS in race_cats
    has_me = defs._race_keys.MIDDLE_EASTERN in race_cats or defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN in race_cats
    has_me_or_sa = defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN in race_cats
    has_pi = defs._race_keys.PACIFIC_ISLANDER in race_cats
    has_multiple = defs._race_keys.MULTIPLE in race_cats
    has_south_asian = defs._race_keys.SOUTH_ASIAN in race_cats or defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN in race_cats
    has_latino = defs._race_keys.LATINO in race_cats
    has_white = defs._race_keys.WHITE in race_cats

    orig = x

    if ((not isinstance(x, str) and pd.notnull(x)) or (isinstance(x, str) and x.isdigit())) and source_name in ["California", "Lincoln"]:
        # Translate datasets with numeric codes for race
        default_cats = defs.get_race_cats(expand=True)
        if source_name == "California":
            # California stops data has a dictionary that can be applied. 
            # https://data-openjustice.doj.ca.gov/sites/default/files/dataset/2023-01/RIPA%20Dataset%20Read%20Me%202021%20Final%20rev%20011223.pdf
            map_dict = {
                1 : default_cats[defs._race_keys.ASIAN], 2 : default_cats[defs._race_keys.BLACK], 3 : default_cats[defs._race_keys.LATINO], 
                4 : default_cats[defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN], 5 : default_cats[defs._race_keys.INDIGENOUS],
                6 : default_cats[defs._race_keys.PACIFIC_ISLANDER], 7 : default_cats[defs._race_keys.WHITE],
                8 : default_cats[defs._race_keys.MULTIPLE]
            }
        elif source_name == "Lincoln":
            # Lincoln data has a dictionary that can be applied.
            # https://ago-item-storage.s3.us-east-1.amazonaws.com/df19cb240d984564965969c568b855c6/LPD_code_tables.pdf?X-Amz-Security-Token=IQoJb3JpZ2luX2VjEFQaCXVzLWVhc3QtMSJHMEUCIQDzuBic65OoJxPSf0DueLmXhFj5uFdAe98fepapTY091AIgLoXQrkkBdRbDjEuwvyGQjpJJLUFgi3ygI8BtjfWuT%2BAq1QQI3f%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FARAAGgw2MDQ3NTgxMDI2NjUiDMxtFJ8yzxImKjHOuSqpBGW%2FA1aH%2F%2FKSYJB1edlTnRu3JE4fzkqCbDBWZrWyKxGmnGWPqnY0hIhhMNS4GSpuwkDUkZANDRh1DUPqVlDZtVfez2tNZwJuCuisqAyofRFAYx%2FQNn18FsJcnXyOZ4hbnduu78jGF4yJU1faAxd1hGrWaKC4NWpYsXxupkdHT5ZDf9SuVGMz8I5L1hB3QLK8cjnxfhZbsXtU6EgBx14jyOdLzrzHBwOFzAkpz7SQJZofNtpRVyRY26xh8RSNL%2Fj1d9LcxjgxchTbiKedc7NZMfoXBWPdb8z%2F3RJMx%2FHVrc6a4NCGc8toWHypuH7lMg1JGxouo9cAc7JxkQrhF9ruPVU54y6EQcxw%2B1JjNfsezqNnk1WLKJ38FJ2zFstJFMfQWNTAqqLk63VjQ2TnnNModiOxlfWLJLfoLSviSqbCVwyQJDzQR6cplKGyICLbIasFPkk%2Fihh1xdH2HeC153E8stesc9FaX5yN09zSYP61TuNOXQicSD9uCAMPEuClxUc0D17iuC%2F9Kxs8Hs1G%2B5C6kD4FBEPYHlAQE4t4nN%2Fehnv7POOgQpYPVo%2Bi%2FI8RRG%2BLPaVI5PQP7ANBQ5ljqqpaIV%2BN0wvt9BdkYi1fsOEsTgg9N2IuN8EdfAXFTFeT%2FmXAZvfEFocvdQqPsV9cXY2cPE3ryEAH0KHzBo6%2Bpfnsay9H2RE23pW%2FvfuAND%2BktpTy0HBzogihKOGs7h4SJBvxUJjTDDBB5agDS6Mwu4SlnwY6qQHNEZx4%2Fa3r4RWoqz%2F%2FC9hkEGXmQNLZZB%2BlcGoE9iVTNLaqK0Wr6oKdgC1xwm5Hq0aUnatarAR5zL8TsG4JkzaCDCggx8Q36CzDsC0sjx1yaEKb4UC6g2MOpY%2Bc0NrIbL%2BQkByyqyreP6797Rx1lHrIvWOrlXaqoB4lMz3Kxxkqe20CCVE5UkTLCAC1X%2Fe5IqLIOzmgigv8tD5vLDE6FaCq3GXMUA%2BUifE3&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20230212T204254Z&X-Amz-SignedHeaders=host&X-Amz-Expires=300&X-Amz-Credential=ASIAYZTTEKKEYIVNVE33%2F20230212%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Signature=da74601aabea63766444b4260baaaec0697753cf9cddf8d70b37204b5126687b
            map_dict = {
                1 : default_cats[defs._race_keys.WHITE], 2 : default_cats[defs._race_keys.BLACK], 3 : default_cats[defs._race_keys.LATINO], 
                4 : default_cats[defs._race_keys.ASIAN], 5 : default_cats[defs._race_keys.INDIGENOUS],
                6 : default_cats[defs._race_keys.OTHER]
            }

        x = int(x)
        if x not in map_dict:
            if no_id=="error":
                raise KeyError(f"Unknown race value for {source_name} data: {orig}")
            else:
                return orig if no_id=="keep" else ""
        else:
            # Replace numerical code with default value
            x = map_dict[x]

    if pd.isnull(x):
        if has_unspecified:
            return race_cats[defs._race_keys.UNSPECIFIED]
        elif no_id=="error":
            raise ValueError(f"Null value found in race column: {orig}")
        else:
            # Same result whether no_id is keep or null
            return ""
                 
    x = x.strip().upper()

    if source_name in ["Austin", "Bloomington", "New York City", "St. John", "Louisville", "Charleston", 
                        "Los Angeles", "Dallas","Chicago"]:
        # Handling dataset-specific letter codes
        if source_name=="Austin":
            # Per email with Kruemcke, Adrian <Adrian.Kruemcke@austintexas.gov> on 2022-06-17
            map_dict = {"M":"Middle Eastern", "P":"Pacific Islander/Native Hawaiian", "N":"Native American/Alaskan", "O":"Other"}
        elif source_name=="Bloomington":
            # https://data.bloomington.in.gov/Police/Citizen-Complaints/kit3-8bua/about_data
            map_dict = {
                "A": "Asian/Pacific Island, Non-Hispanic", 
                "B": "African American, Non-Hispanic", 
                "C": "Hawaiian/Other Pacific Island, Hispanic", 
                "H": "Hawaiian/Other Pacific Island, Non-Hispanic", 
                "I": "Indian/Alaskan Native, Non-Hispanic", 
                "K": "African American, Hispanic", 
                "L": "Caucasian, Hispanic", 
                "N": "Indian/Alaskan Native, Hispanic", 
                "P": "Asian/Pacific Island, Hispanic", 
                "S": "Asian, Non-Hispanic", 
                "T": "Asian, Hispanic", 
                "U": "Unknown", 
                "W": "Caucasian, Non-Hispanic"
            } 
        elif source_name == "New York City":
            # https://www.nyc.gov/assets/nypd/downloads/zip/analysis_and_planning/stop-question-frisk/SQF-File-Documentation.zip
            map_dict = {"P":"Black-Hispanic", "Q":"White-Hispanic", "X":"Unknown","Z":"Other"}
        elif source_name == "St. John":
            # https://stjohnin.gov/PD/PDI/Warnings/2019Warnings.php
            map_dict = {"L":"White/Hispanic/Latin","M":"Multiracial","N":"Asian Indian"}
        elif source_name == "Louisville":
            map_dict = {"A":"Asian/Pacific Islander", "U":"Undeclared", "IB":"Indian/India/Burmese", "M":"Middle Eastern Descent","AN":"Alaskan Native"}
        elif source_name=="Charleston":
            # Based on correspondance with City of Charleston Ombudsman
            map_dict = {"A":"Asian or Pacific Islander", "AI": "Alaskan or American Indian", "AP":"Asian or Pacific Islander",
                        "BK":"Black","MR":"Multi-Racial","AO":"Other"}
        elif source_name=="Los Angeles":
            # https://data.lacity.org/Public-Safety/Traffic-Collision-Data-from-2010-to-Present/d5tf-ez2w
            map_dict = {"A":"Other Asian", "B":"Black", "C":"Chinese", "D":"Cambodian", "F":"Filipino", 
                        "G":"Guamanian", "H":"Hispanic/Latin/Mexican", "I":"American Indian/Alaskan Native", 
                        "J":"Japanese", "K":"Korean", "L":"Laotian", "O":"Other", "P":"Pacific Islander", 
                        "S":"Samoan", "U":"Hawaiian", "V":"Vietnamese", "W":"White", "X":"Unknown", "Z":"Asian Indian"}
        elif source_name=="Dallas":
            # Based on Bloomington and St. John usage as well as names associated with usage
            map_dict = {"L":"Caucasian, Hispanic"}
        elif source_name=="Chicago":
            # https://home.chicagopolice.org/wp-content/uploads/2020/08/ISR-Data-Dictionary.csv
            map_dict = {
                'BLK' : 'BLACK',
                'WHI' : 'WHITE',
                'API' : 'ASIAN/PACIFIC ISLANDER',
                'WBH' : 'BLACK HISPANIC',
                'WWH' : 'WHITE HISPANIC',
                'I' : 'AMER IND/ALASKAN NATIVE',
                'U' : 'UNKNOWN',
                'P' : 'NATIVE HAWAIIAN OR OTHER PACIFIC ISLANDER',
                'WHT' : 'WHITE'
            }

        if x in map_dict:
            x = map_dict[x].upper()

    if _p_nonlatino.search(x) and len(m:=_p_nonlatino.sub('', x))>0:
        x = m

    if isinstance(x,str):
        # Look for code: {Race_Abbreviation_Initial} {- or =} {Full name}
        abbrev_full_match = re.search(r"^([\w\s/\.]+)\s?[-=]\s?([\w\s/\.]+)$",x)
        if abbrev_full_match and any([len(x)==1 for x in abbrev_full_match.groups()]):
            x = [x for x in abbrev_full_match.groups() if len(x)>1][0].strip()

        delims = [" and ", ",",'|','/',';']
        delim = [d for d in delims if d in x]
        # Look for delimited multi-race data
        if not known_single and len(delim)>0 and \
            not x.lower().startswith("some other") and \
            x.lower().replace(" ","") not in ["hawaiian/pacific", "middleeastern/southasian", 'n/a',
                                              'asian/pacis','unk/oth','oth/unk', 'black/africanamerican','hispanic/latino',
                                              "americanindian/alaskanative",'a/indian'] and \
            not any([x.lower() in ['unknown','other'] for x in x.split(delim[0])]):
            # Treat this as a list of races
            delim = delim[0]
            race_list = []
            for v in x.split(delim):
                if (v=="INDIAN" and "BURMESE" in x) or v=='N':  # N is non-Latino, which can be ignored
                    # Handle special case to prevent setting someone from country of India to Indigenous
                    continue
                new_val = _create_race_lut(v, no_id, source_name, race_cats, agg_cat, known_single=True)
                if isinstance(new_val, list):
                    race_list.extend(new_val)
                else:
                    race_list.append(new_val)

            if has_unspecified and agg_cat:
                num_unspecified = sum([x==race_cats[defs._race_keys.UNSPECIFIED] for x in race_list])
                if num_unspecified == len(race_list):
                    # All are unspecified. Just return unspecified rather than a list
                    return race_cats[defs._race_keys.UNSPECIFIED]
                else:
                    # Ignore the unspecifieds
                    race_list = [x for x in race_list if x!=race_cats[defs._race_keys.UNSPECIFIED]]

            if len(race_list)==1 or all([x==race_list[0] for x in race_list]): # All are the same
                return race_list[0]
            elif has_aapi and has_asian and has_pi and \
                set(race_list) == set([race_cats[defs._race_keys.ASIAN], race_cats[defs._race_keys.PACIFIC_ISLANDER]]):
                # Simplify to AAPI
                return race_cats[defs._race_keys.AAPI]
            elif has_aapi and has_asian and \
                set(race_list) == set([race_cats[defs._race_keys.ASIAN], race_cats[defs._race_keys.AAPI]]):
                # Simplify to AAPI
                return race_cats[defs._race_keys.AAPI]
            elif agg_cat and has_latino and race_cats[defs._race_keys.LATINO] in race_list:
                # Category is Latino of all races
                return race_cats[defs._race_keys.LATINO]
            else:
                return race_list

    # Clean x
    x = x.upper().replace("_", " ").replace("*","").replace("-","").replace(".","").strip()

    if pd.notnull(x):
        # Check if value is part of race_cats
        # This is particularly necessary for custom race_cats dicts
        for k,v in race_cats.items():
            # Key is always be a string
            if k.upper()==str(orig).upper() or k.upper()==str(x) or str(v)==str(orig).upper() or str(v)==str(x):
                return v
        
    def is_latino(x):
        if isinstance(x,str):
            x = x.upper().replace("-","").replace(" ","")
            return "HISPANIC" in x or "LATINO" in x
        else:
            return False

    x_no_space = x.replace(" ","")
    if has_unspecified and (x in ["MISSING","NOT SPECIFIED", "", "NOT RECORDED","N/A", "NOT REPORTED", "NONE"] or \
        (type(x)==str and ("NO DATA" in x or ("NOT APP" in x and x not in ["NOT APPLICABLE (NONUS)",'NOT APPLICABLE (NON US)']) or "NO RACE" in x or "NULL" in x))):
        return race_cats[defs._race_keys.UNSPECIFIED]
    if has_white and x_no_space in ["W", "CAUCASIAN", "WN", "WHITE", "WHTE", "WHITE,OTHER", 'WHT']:  # WN = White-Non-Hispanic
        return race_cats[defs._race_keys.WHITE]
    if has_black and (x in ["B", "AFRICAN AMERICAN", "BLCK", "BLK", "BLACE",'AFR AMERICAN'] or \
                      x_no_space in ["AFRICANAMERICAN"] or \
                      re.search("BLACK?($|[^A-Za-z])",x)) \
                      and not is_latino(x):
        if x.count("BLACK") > 1:
            raise ValueError(f"The value of {x} likely contains the races for multiple people")
        return race_cats[defs._race_keys.BLACK]
    if has_south_asian and ((x in ["SOUTH ASIAN"] or ("ASIAN" in x and "INDIAN" in x)) or \
                            x in ["EAST INDIAN"]):
        if defs._race_keys.SOUTH_ASIAN in race_cats:
             return race_cats[defs._race_keys.SOUTH_ASIAN]
        else:
            return race_cats[defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN] 
    if has_me and (x in ["ME","ARABIC"] or "MIDDLE EAST" in x) and "AFRICA" not in x:  # Ignore Middle Eastern or North African
        if "SOUTH ASIAN" in x:
            if has_me_or_sa:
                return race_cats[defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN]
        else:
            return race_cats[defs._race_keys.MIDDLE_EASTERN] if defs._race_keys.MIDDLE_EASTERN in race_cats else race_cats[defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN]
    if (has_asian or has_aapi) and (x in ["A", 'ORIENTAL', 'AA', 'ASN'] or "ASIAN" in x.replace("CAUCASIAN","")) and x not in ["SOUTHWEST ASIAN"] and "INDIAN" not in x:
        if has_aapi and ("PAC" in x or "HAWAI" in x):
            return race_cats[defs._race_keys.AAPI] 
        else:
            return race_cats[defs._race_keys.ASIAN] if has_asian else race_cats[defs._race_keys.AAPI]
    if (has_pi or has_aapi) and ("HAWAI" in x or "PACIFIC" in x_no_space or \
                                 "PACISL" in x_no_space or x in ['PI', 'NHPI']): # NHPI = Native Hawaiian, and Pacific Islander
        return race_cats[defs._race_keys.PACIFIC_ISLANDER] if has_pi else race_cats[defs._race_keys.AAPI]
    if has_latino and x in ["H", "WH", "HISPANIC", "LATINO", "HISPANIC OR LATINO", "LATINO OR HISPANIC", 
                            "HISPANIC/LATINO", "LATINO/HISPANIC",'HISPANIC/LATIN/MEXICAN','HISP']:
        return race_cats[defs._race_keys.LATINO] 
    if has_indigenous and (x in ["I", "INDIAN", "ALASKAN NATIVE", "AN", "AI", "AL NATIVE","A/INDIAN", "NAT AM"] or "AMERICAN IND" in x.replace("II","I") or \
        "NATIVE AM" in x or  "AMERIND" in x_no_space or "ALASK" in x or "AMIND" in x_no_space or \
        "NAT AMER" in x):
        return race_cats[defs._race_keys.INDIGENOUS] 
    if has_multiple and ("OR MORE" in x or "MULTI" in x or \
        x_no_space in ["MIXED","BIRACIAL","MIXEDRACE","MORE THAN TWO RACES".replace(" ",""),
                              "MORE THAN 2 RACES".replace(" ","")]):
        return race_cats[defs._race_keys.MULTIPLE]
    if defs._race_keys.OTHER_UNKNOWN in race_cats and "UNK" in x and "OTH" in x:
        return race_cats[defs._race_keys.OTHER_UNKNOWN]
    if defs._race_keys.UNKNOWN in race_cats and  ("UNK" in x or x in ["U", "UK"]):
        return race_cats[defs._race_keys.UNKNOWN]
    if defs._race_keys.OTHER in race_cats and (x in ["O","OTHER","OTH"] or "OTHER UNCLASS" in x or "OTHER RACE" in x):
        return race_cats[defs._race_keys.OTHER]
    
    if agg_cat:
        # This attempts to categorize more cases based on U.S. Census definitions:
        # https://www.census.gov/programs-surveys/decennial-census/decade/2020/planning-management/release/faqs-race-ethnicity.html
        # Cases include here are based on data values that have been observed
        if has_latino and (("HISP" in x and "NONHISP" not in x_no_space) or \
                           ("LATIN" in x and "NONLATINO" not in x_no_space) or \
                            x in ["MEXICAN"]):
            # agg_cat forces Latino of all races
            return race_cats[defs._race_keys.LATINO] 
        elif has_black and x in ["EAST AFRICAN"]:
            return race_cats[defs._race_keys.BLACK]
        elif has_white and x in ["BOSNIAN"]:
            return race_cats[defs._race_keys.WHITE]
        elif has_me and x in ["SOUTHWEST ASIAN"]:
            return race_cats[defs._race_keys.MIDDLE_EASTERN] if defs._race_keys.MIDDLE_EASTERN in race_cats else race_cats[defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN]
        elif has_south_asian and (("INDIAN" in x and "BURMESE" in x) or x in ["INDIA","BURMESE"]):
            if defs._race_keys.SOUTH_ASIAN in race_cats:
                return race_cats[defs._race_keys.SOUTH_ASIAN]
            else:
                return race_cats[defs._race_keys.MIDDLE_EASTERN_SOUTH_ASIAN]
        elif (has_asian or has_aapi) and (x in ["CAMBODIAN",'VIETNAMESE',"LAOTIAN","JAPANESE","KOREAN","CHINESE","HMONG","MIEN","FILIPINO"] or "ASIAN" in x):
            return race_cats[defs._race_keys.ASIAN] if has_asian else race_cats[defs._race_keys.AAPI]
        elif (has_pi or has_aapi) and x in ["POLYNESIAN","SAMOAN","GUAMANIAN"]:
            return race_cats[defs._race_keys.PACIFIC_ISLANDER] if has_pi else race_cats[defs._race_keys.AAPI]
        elif no_id=="error":
            raise ValueError(f"Unknown value in race column: {orig}")
        elif no_id=="test":
            if x in ["MALE","FEMALE","GIVING ANYTHING OF VALUE","REFUSED", "NA","M","F","OTHER/NOT REPORTED", 
                     'UNOCCUPIED VEHICLE','MIDDLE EASTERN OR NORTH AFRICAN','P','PAKISTANI','DUPLICATE',
                     'NOT AVAILABLE'] or \
                (source_name in ["Chapel Hill","Lansing","Fayetteville"] and x in ["S","P"]) or \
                (source_name=="Burlington" and x in ["EXPUNGED"]) or \
                (source_name=="Minneapolis" and x in ["NO CONTACT"]) or \
                (source_name in ["Cincinnati","San Diego"] and x in ["F","S","P"]) or \
                (source_name in ["Columbia"] and x in ["M","P",'34']) or \
                (source_name in ["Urbana"] and x in ["BUSINESS OR OTHER"]) or \
                (source_name in ["Bloomington","Beloit"] and x in ["M"]) or \
                (source_name in ["Beloit"] and x in ["L"]) or \
                (source_name in ["St. John"] and x in ["K","P"]) or \
                (source_name in ["Rutland"] and x in ["M","R"]) or \
                (source_name in ["Dallas"] and x in ["NA"]) or \
                (source_name in ["Sacramento"] and x in ["CUBAN","CARRIBEAN"]) or \
                (source_name in ["New York City"] and x in ["SOUTHWEST"]) or \
                (source_name in ["Wallkill"] and x in ["IND",'N','M']) or \
                (x=="UN" and source_name=="State Police") or \
                (x=="P" and source_name=="Pittsfield") or \
                (x=="PENDING RELEASE" and source_name=="Portland") or \
                (x=="IRAK" and source_name=="Chattanooga") or \
                (x=="N" and source_name=="Bremerton") or \
                (x=="W\nW" and source_name=="Sparks") or \
                (x in ['UI'] and source_name=="Norwich") or \
                (x in ['E','P'] and source_name=="Tucson") or \
                ("DOG" in x) or \
                (source_name in ["New Orleans"] and "NOT APPLICABLE (NON" in x) or \
                (source_name in ["Detroit", "Fairfax County"] and x in ["N","SELECT","UNVERIFIED",'240','160','180','P','120']) or \
                x in ["OTHER / MIXED RACE", "UNDISCLOSED", "OR SPANISH ORIGIN","PREFER NOT TO SAY","OTHERBLEND","UNDECLARED"] or \
                len(orig)>100:
                # This is meant to be temporary for testing
                return "BAD DATA"
            elif "EXEMPT" in x:
                return orig
            else:
                raise ValueError(f"Unknown value in race column: {orig}")
        else:
            return orig if no_id=="keep" else ""
    elif no_id=="error":
        raise ValueError(f"Unknown value in race column: {orig}")
    else:
        return orig if no_id=="keep" else ""


def _create_gender_lut(x, no_id, source_name, gender_cats, *args, **kwargs):

    orig = x
    if pd.notnull(x) and (type(x) != str or x.isdigit()) and source_name in ["California", "Lincoln"]:
        default_cats = defs.get_gender_cats()
        if source_name == "California":
            # California stops data has a dictionary that can be applied. 
            # https://data-openjustice.doj.ca.gov/sites/default/files/dataset/2023-01/RIPA%20Dataset%20Read%20Me%202021%20Final%20rev%20011223.pdf
            map_dict = {
                1 : default_cats[defs._gender_keys.MALE], 2 : default_cats[defs._gender_keys.FEMALE], 3 : default_cats[defs._gender_keys.TRANSGENDER_MALE], 
                4 : default_cats[defs._gender_keys.TRANSGENDER_FEMALE], 5 : default_cats[defs._gender_keys.GENDER_NONCONFORMING]
            }
        elif source_name == "Lincoln":
            # Lincoln data has a dictionary that can be applied.
            # https://ago-item-storage.s3.us-east-1.amazonaws.com/df19cb240d984564965969c568b855c6/LPD_code_tables.pdf?X-Amz-Security-Token=IQoJb3JpZ2luX2VjEFQaCXVzLWVhc3QtMSJHMEUCIQDzuBic65OoJxPSf0DueLmXhFj5uFdAe98fepapTY091AIgLoXQrkkBdRbDjEuwvyGQjpJJLUFgi3ygI8BtjfWuT%2BAq1QQI3f%2F%2F%2F%2F%2F%2F%2F%2F%2F%2FARAAGgw2MDQ3NTgxMDI2NjUiDMxtFJ8yzxImKjHOuSqpBGW%2FA1aH%2F%2FKSYJB1edlTnRu3JE4fzkqCbDBWZrWyKxGmnGWPqnY0hIhhMNS4GSpuwkDUkZANDRh1DUPqVlDZtVfez2tNZwJuCuisqAyofRFAYx%2FQNn18FsJcnXyOZ4hbnduu78jGF4yJU1faAxd1hGrWaKC4NWpYsXxupkdHT5ZDf9SuVGMz8I5L1hB3QLK8cjnxfhZbsXtU6EgBx14jyOdLzrzHBwOFzAkpz7SQJZofNtpRVyRY26xh8RSNL%2Fj1d9LcxjgxchTbiKedc7NZMfoXBWPdb8z%2F3RJMx%2FHVrc6a4NCGc8toWHypuH7lMg1JGxouo9cAc7JxkQrhF9ruPVU54y6EQcxw%2B1JjNfsezqNnk1WLKJ38FJ2zFstJFMfQWNTAqqLk63VjQ2TnnNModiOxlfWLJLfoLSviSqbCVwyQJDzQR6cplKGyICLbIasFPkk%2Fihh1xdH2HeC153E8stesc9FaX5yN09zSYP61TuNOXQicSD9uCAMPEuClxUc0D17iuC%2F9Kxs8Hs1G%2B5C6kD4FBEPYHlAQE4t4nN%2Fehnv7POOgQpYPVo%2Bi%2FI8RRG%2BLPaVI5PQP7ANBQ5ljqqpaIV%2BN0wvt9BdkYi1fsOEsTgg9N2IuN8EdfAXFTFeT%2FmXAZvfEFocvdQqPsV9cXY2cPE3ryEAH0KHzBo6%2Bpfnsay9H2RE23pW%2FvfuAND%2BktpTy0HBzogihKOGs7h4SJBvxUJjTDDBB5agDS6Mwu4SlnwY6qQHNEZx4%2Fa3r4RWoqz%2F%2FC9hkEGXmQNLZZB%2BlcGoE9iVTNLaqK0Wr6oKdgC1xwm5Hq0aUnatarAR5zL8TsG4JkzaCDCggx8Q36CzDsC0sjx1yaEKb4UC6g2MOpY%2Bc0NrIbL%2BQkByyqyreP6797Rx1lHrIvWOrlXaqoB4lMz3Kxxkqe20CCVE5UkTLCAC1X%2Fe5IqLIOzmgigv8tD5vLDE6FaCq3GXMUA%2BUifE3&X-Amz-Algorithm=AWS4-HMAC-SHA256&X-Amz-Date=20230212T204254Z&X-Amz-SignedHeaders=host&X-Amz-Expires=300&X-Amz-Credential=ASIAYZTTEKKEYIVNVE33%2F20230212%2Fus-east-1%2Fs3%2Faws4_request&X-Amz-Signature=da74601aabea63766444b4260baaaec0697753cf9cddf8d70b37204b5126687b
            map_dict = {
                 1 : default_cats[defs._gender_keys.MALE], 2 : default_cats[defs._gender_keys.FEMALE]
            }

        x = int(x)
        if x not in map_dict:
            if no_id=="error":
                raise KeyError(f"Unknown gender value for {source_name} data: {orig}")
            else:
                return orig if no_id=="keep" else ""
        else:
            # Replace numerical code with default value
            x = map_dict[x]


    if isinstance(x,str):
        # Look for {Abbreviation_Initial} {- or =} {Full name}
        abbrev_full_match = re.search(r"^([\w\s/\.]+)\s?[-=]\s?([\w\s/\.]+)$",x)
        if abbrev_full_match and any([len(x)==1 for x in abbrev_full_match.groups()]):
            x = [x for x in abbrev_full_match.groups() if len(x)>1][0].strip()
        x = x.upper().replace("-","").replace("_","").replace(" ","").replace("'","")

        if source_name in ["New York City", "Los Angeles", "Chicago",'New York']:
            # Handling dataset-specific codes
            if source_name == "New York City":
                # https://www.nyc.gov/assets/nypd/downloads/zip/analysis_and_planning/stop-question-frisk/SQF-File-Documentation.zip
                map_dict = {"Z":"Unknown"}
            elif source_name == "New York":
                # https://data.ny.gov/Transportation/Traffic-Tickets-Issued-Four-Year-Window/q4hy-kbtf/about_data
                map_dict = {"C":"Organization"}
            elif source_name in ["Los Angeles",'Chicago']:
                # https://data.lacity.org/Public-Safety/Traffic-Collision-Data-from-2010-to-Present/d5tf-ez2w
                # https://home.chicagopolice.org/wp-content/uploads/2020/08/ISR-Data-Dictionary.csv
                map_dict = {"X":"Unknown"}

            if x.upper() in map_dict:
                x = map_dict[x.upper()].upper()

    if pd.notnull(x):
        # Check if value is part of gender_cats
        # This is particularly necessary for custom gender_cats dicts
        for k,v in gender_cats.items():
            # Key is always be a string
            if k.upper()==str(orig).upper() or k.upper()==str(x) or str(v)==str(orig).upper() or str(v)==str(x):
                return v

    has_unspecified = defs._gender_keys.UNSPECIFIED in gender_cats
    if pd.isnull(x):
        if has_unspecified:
            return gender_cats[defs._gender_keys.UNSPECIFIED]
        elif no_id=="error":
            raise ValueError(f"Null value found in gender column: {orig}")
        else:
            # Same result whether no_id is keep or null
            return ""
    if has_unspecified and x in ["MISSING", "UNSPECIFIED", "",",",'NOTSPECIFIED',"NOTRECORDED","NONE"] or \
        "NODATA" in x or "NOSEX" in x or "NULL" in x:
        return gender_cats[defs._gender_keys.UNSPECIFIED]
    elif defs._gender_keys.FEMALE in gender_cats and x in ["F", "FEMALE", "FEMAALE", "FFEMALE", "FEMAL", "FEMALE/WOMAN","WOMAN"]:
        return gender_cats[defs._gender_keys.FEMALE]
    elif defs._gender_keys.MALE in gender_cats and x in ["M", "MALE", "MMALE", "MALE/MAN","MAN", "MLE"]:
        return gender_cats[defs._gender_keys.MALE]
    elif defs._gender_keys.OTHER in gender_cats and x in ["OTHER", "O"]:
        return gender_cats[defs._gender_keys.OTHER]
    elif defs._gender_keys.TRANSGENDER in gender_cats and x in ["TRANSGENDER","TRANSGENDERED"]:
        return gender_cats[defs._gender_keys.TRANSGENDER]
    elif defs._gender_keys.GENDER_NONBINARY in gender_cats and x in ["NONBINARY"]:
        return gender_cats[defs._gender_keys.GENDER_NONBINARY]
    elif defs._gender_keys.TRANSGENDER_OR_GENDER_NONCONFORMING in gender_cats and (x in [
        "Gender Diverse (gender non-conforming and/or transgender)".upper().replace("-","").replace(" ",""),
        "GENDERNONCONFORMING",'NONCONFORMING','GENDERDIVERSE'] or "TGNC" in x):
        return gender_cats[defs._gender_keys.TRANSGENDER_OR_GENDER_NONCONFORMING]
    elif defs._gender_keys.TRANSGENDER_MALE in gender_cats and (x in ["TRANSGENDER MALE".replace(" ","")] or \
        "TRANSGENDERMAN" in x or \
        (source_name=="Los Angeles" and x=="T")):
        return gender_cats[defs._gender_keys.TRANSGENDER_MALE]
    elif defs._gender_keys.TRANSGENDER_FEMALE in gender_cats and \
        (x in ["TRANSGENDER FEMALE".replace(" ","")] or \
        "TRANSGENDERWOMAN" in x or \
        (source_name=="Los Angeles" and x=="W")):
        return gender_cats[defs._gender_keys.TRANSGENDER_FEMALE]
    elif defs._gender_keys.GENDER_NONCONFORMING in gender_cats and (source_name=="Los Angeles" and x=="C"):
        return gender_cats[defs._gender_keys.GENDER_NONCONFORMING]
    elif defs._gender_keys.UNKNOWN in gender_cats and \
        (x in ["U","UK", "UNABLE TO DETERMINE".replace(" ","")] or "UNK" in x):
        return gender_cats[defs._gender_keys.UNKNOWN]
    elif defs._gender_keys.UNSPECIFIED_OR_ANOTHER in gender_cats and x in 'X':
        # https://www.state.gov/x-gender-marker-available-on-u-s-passports-starting-april-11/
        return gender_cats[defs._gender_keys.UNSPECIFIED_OR_ANOTHER]
    
    if no_id=="test":
        bad_data = ["BLACK","WHITE",'HISPANIC','ASIAN','FEMALE10','DUPLICATE']
        if "EXEMPT" in x or x in ["DATAPENDING", "NOTAPPLICABLE","N/A",'NA','NOTAVAILABLE'] or ("BUSINESS" in x and source_name=="Cincinnati"):
            return orig
        elif x in bad_data or \
            (source_name=="New York City" and (x=="Z" or x.isdigit())) or \
            (source_name=="New York" and x in ["Organization".upper()]) or \
            (source_name=="Baltimore" and x in ["Y","Z"]) or \
            (source_name=="Urbana" and x in ["."]) or \
            (source_name=="Greensboro" and x in ["ASIAN"]) or \
            (source_name=="Columbia" and x in ["B"]) or \
            (source_name=="Burlington" and x in ["EXPUNGED"]) or \
            (source_name=="Fairfax County" and x in ['O']) or \
            (source_name=='Albemarle County' and x in ['N']) or \
            (source_name in ["Seattle","New Orleans","Menlo Park","Rutland"] and x in ["D","N"]) or \
            (source_name=="Los Angeles County" and x=="0") or \
            (x in ["W","NA"]) or \
            (x in ['L'] and source_name=='Bloomington') or \
            (x in ['GROUP'] and source_name=='Asheville') or \
            (x in ['UI'] and source_name=="Norwich") or \
            (x=="MA" and source_name in ["Lincoln"]) or \
            (x=="P" and source_name=="Fayetteville") or \
            (x in ['B'] and source_name=="Louisville") or \
            (x=="5" and source_name=="Lincoln") or \
            (source_name=="Minneapolis" and x in ["NOCONTACT"]) or \
            (x in ["MALE,MALE"] and source_name=="Chattanooga") or \
            (x=="PENDINGRELEASE" and source_name=="Portland") or \
            (x in ["N","H"] and source_name=="Los Angeles") or \
            (source_name=='Bremerton' and x in ['B', 'W', 'A','I']) or \
            "DOG" in x or \
            x in ["UNDISCLOSE","UNDISCLOSED","PREFER TO SELF DESCRIBE".replace(" ",""),'NONBINARY/THIRDGENDER',
                  "PREFER NOT TO SAY".replace(" ",""),"TGNC/OTHER","REFUSED",'UNVERIFIED','NONBINARY/OTHERX','UNOCCUPIEDVEHICLE'] or \
            source_name == "Buffalo":
            return orig
        else:
            raise ValueError(f"Unknown value in gender column: {orig}")
    if no_id=="error":
        raise ValueError(f"Unknown value in gender column: {orig}")
    else:
        return orig if no_id=="keep" else ""
    

def _create_injury_lut(x, no_id, source_name, cats, *args, **kwargs):
    if pd.isnull(x):
        return "UNSPECIFIED"
    elif isinstance(x,Number) or pd.api.types.is_bool(x) or x.isdigit():  # is_bool handles numpy.bool_ type
        x = int(x)
        if x<0:
            return x
        else:
            return "INJURED" if x>0 else "NO INJURY"
    orig = x
    x = re.sub(r'^\d+\s*\-\s*','',x)
    x = x.upper().replace('-',' ').replace('*','').replace('SUBJECT','').replace('  ',' ').strip()
    x = re.sub(r'OF[FI]{2}CER','', x).strip()  # Handle misspelling officer
    x = re.sub(r'\s{2,10}',' ',x)

    if len(x)==0:
        return "UNSPECIFIED"
    elif len(words:=x.split('\n'))>1:
        words = [_create_injury_lut(y, no_id, source_name, cats) for y in words]
        for m in ['FATAL','INJURED']:
            if m in words:
                return m

    contains_yes = x.replace(',',' ').startswith('YES ')
    nonfatal_words = ["INJURED", 'NON FATAL', 'NON FATAL','NON FATAL INJURY', 'INJURY'] 
    contains_nonfatal = any([y in x for y in nonfatal_words])
    fatal_words = ["FATAL","KILLED",'DECEASED',"DEATH", 'FATAL INJURY']
    contains_fatal = any([y in x for y in fatal_words])
    is_fatal = contains_yes and contains_fatal and not contains_nonfatal

    if is_fatal or x in fatal_words:
        return "FATAL"
    elif x.startswith('NO INJUR') or x.startswith('NONE') or x.startswith('NO COMPLAINT') or\
        x in ["NOT INJURED",'NEITHER','NO','N', "MISS", 'SHOOT AND MISS','FALSE','NO VISIBLE INJURY','UNINJURED', 'SHOW OF FORCE']:
        return "NO INJURY"
    elif contains_yes or x in nonfatal_words or x in ['Y','YES','TRUE'] or \
        any([x.startswith(y) for y in ['COMPLAINED OF','COMPLAINT OF']]) or \
        any([y in x for y in ['WOUND','PAIN', "BLEEDING",'SWELLING','SCRAPE','PUNCTURE','LACERATION','BRUIS',
                              'BROKEN','UNCONSCIOUS', 'FIRST AID', 'SHOT (INJURED ONLY)', 'INJURED INCIDENTALLY',
                              'DISLOCATED','FRACTURED','ABBRASION','ABRASION','ABRAISON','ABRASSION','BUSTED','PULLED OUT','REDNESS','LOSS','RASH',
                              'SCRATCH','NUMBNESS','BREATHING','CUT','STUN', 'MARKS', 'EYE', 'PEELING', 'HURT', 'ELBOW', 'KNEE',
                              'SOFT TISSUE','BLOOD','HEAD','SORE','SHOULDER', 'MINOR INJUR', 'FINGER', 'IMPACT', 'FACE', 'ARM',
                              'MOUTH', 'BACK','RIB', 'THUMB','SHIN',' EAR', 'ACHILLES', 'STRUCK', 'LEG', 'SERIOUS',
                              'CONCUSSION','FRACTURE','CANINE BITE', 'MARK','BURN', 'MINOR', 'DISABL', 'PHYSICAL INJURY',
                              'TREATED','TAKEN TO HOSPITAL']]):
        return "INJURED"
    elif 'NALOXONE' in x:
        return orig
    elif x in ['SELF INFLICTED FATAL', 'DECEASED (SELF INFLICTED)','KILLED (SELF INFLICTED)', 'FATAL SUICIDE']:
        return 'SELF-INFLICTED FATAL'
    elif x in ['UNKNOWN','UNKNWON']:
        return 'UNKNOWN'
    elif x in ['UNSPECIFIED', "NOT AVAILABLE", 'NOT SPECIFIED','(MISSING)','NA']:
        return 'UNSPECIFIED'
    elif x=='OTHER':
        return 'OTHER'
    elif no_id=='test':
        if source_name not in ['Indianapolis', 'Northampton','Minneapolis','Rutland'] and \
            (x not in ['UI'] or source_name!="Norwich"):
            raise ValueError(f"Unknown value in injury column: {orig}")
    elif no_id=='error':
        raise ValueError(f"Unknown value in injury column: {orig}")
    else:
        return orig if no_id=="keep" else ""
    
def _create_fatal_lut(x, no_id, source_name, cats, *args, **kwargs):
    if pd.isnull(x):
        return "UNSPECIFIED"
    elif isinstance(x,Number) or x.isdigit():
        x = int(x)
        if x<0:
            return x
        elif x>1:
            if no_id=='error':
                raise ValueError(f"Unknown value in fatal column: {orig}")
            else:
                return orig if no_id=="keep" else ""
        else:
            return "YES" if x>0 else "NO"
    orig = x
    x = x.upper().strip()
    if x in ["FATAL","YES", "Y",'DECEASED']:
        return "YES"
    elif x in ["NON-FATAL","NON FATAL", "NO","N",'NO CONTACT']:
        return "NO"
    elif x in ["SELF-INFLICTED"] or (source_name=='Louisville' and x=='YS'):
        return "SELF-INFLICTED FATAL"
    elif (source_name=='Louisville' and x=='CONTINUED'):
        # Used in Louisville table to indicate where a single shooting applies to multiple officers in different rows
        return orig
    elif no_id in ['test', 'error']:
        raise ValueError(f"Unknown value in fatal column: {orig}")
    else:
        return orig if no_id=="keep" else ""
    

def _create_firearm_lut(x, no_id, source_name, cats, *args, **kwargs):
    if pd.isnull(x):
        return "UNSPECIFIED"
    elif isinstance(x,Number) or x.isdigit():
        x = int(x)
        if x<0:
            return x
        else:
            return "YES" if x>0 else "NO"
    orig = x
    x = x.upper().strip()
    if x in ["YES", "Y",'TRUE','FIREARM']:
        return "YES"
    elif x in ["NO","N",'FALSE','PHYSICAL FORCE','ELECTRICAL WEAPON','OC SPRAY','IMPACT WEAPON', 'NO FORCE'] or \
        any([y in x for y in ['CANINE','BLANKET','DISPLAYED','CUTTING']]):
        return "NO"
    elif no_id=='test':
        raise ValueError(f"Unknown value in injury column: {orig}")
    elif no_id=='error':
        raise ValueError(f"Unknown value in injury column: {orig}")
    else:
        return orig if no_id=="keep" else ""
    

def std_dict(col, std_map, converter, *args):
    # Standardize column containing dictionaries containing the races of multiple people
    new_vals = []
    for j in col.index:
        p_dict = col[j].copy()
        for k, v in p_dict.items():
            if v not in std_map:
                std_map[v] = converter(v, *args)
            p_dict[k] = std_map[v]
        new_vals.append(p_dict)

    # If all values are of length
    if all([len(x)==1 for x in new_vals]):
        # No need for dictionary
        new_vals = [x[0] for x in new_vals]

    return new_vals


def std_demo_col(col, vals, map_dict, item_num, converter, *args):
    # Standardize demographics column
    map = {}
    delims = ["(", ","]
    multi_found = False
    for x in vals:
        if isinstance(x,str):
            items = [x]
            for d in delims:
                if d in x:
                    items = x.split(d)
                    break
            
            new_val = {}
            if len(items[0])==0:
                istart = 1
            else:
                istart = 0
            for k, i in enumerate(items[istart:]):
                if "," in i:
                    i = i.split(",")[item_num].strip()
                elif "/" in i:
                    i = i.split("/")[item_num].strip()
                elif i != "Unk":
                    raise NotImplementedError()

                if i not in map_dict:
                    map_dict[i] = converter(i, *args)

                if len(items[istart:])==1:
                    new_val = map_dict[i]
                else:
                    new_val[k] = map_dict[i]
                    multi_found = True

            map[x] = new_val
        else:
            if x not in map_dict:
                map_dict[x] = converter(x, *args)
                map[x] = map_dict[x]

    if multi_found:
        # Convert non-dicts to dict
        map = {k:(v if (isinstance(v,dict) or pd.isnull(v)) else {0:v}) for k,v in map.items()}

    return col.map(map)


def std_counts(col, vals, map_dict, delim, converter, *args, **kwargs):
    map = {}
    race_count_re = re.compile(r"(\d+)\s?-\s?([A-Za-z]+\.?\s?[A-Za-z]*\.?)")
    for x in vals:
        matches = race_count_re.findall(x) if isinstance(x,str) else False
        if matches:
            total = sum([int(m[0]) for m in matches])
            k = 0
            new_val = {}
            for m in matches:
                if m[1] not in map_dict:
                    map_dict[m[1]] = converter(m[1], *args, **kwargs)

                if total==1:
                    new_val = map_dict[m[1]]
                else:
                    for _ in range(int(m[0])):
                        new_val[k] = map_dict[m[1]]
                        k+=1
                    multi_found = True

            map[x] = new_val
        else:
            if x not in map_dict:
                map_dict[x] = converter(x, *args, **kwargs)
                map[x] = map_dict[x]

    if multi_found:
        # Convert non-dicts to dict
        map = {k:(v if (isinstance(v,dict) or pd.isnull(v)) else {0:v}) for k,v in map.items()}

    return col.map(map)


def std_with_names(col, vals, map_dict, item_num, converter, *args, **kwargs):
    map = {}
    multi_found = False
    p = re.compile(r"[\sÊ](\w{1,2}/\w)")
    for x in vals:
        if isinstance(x,str):
            items = p.findall(x)
            
            new_val = {}
            for k, i in enumerate(items):
                i = i.split("/")
                i = i[item_num]

                if i not in map_dict:
                    map_dict[i] = converter(i, *args)

                if len(items)==1:
                    new_val = map_dict[i]
                else:
                    new_val[k] = map_dict[i]
                    multi_found = True

            map[x] = new_val
        else:
            if x not in map_dict:
                map_dict[x] = converter(x, *args)
                map[x] = map_dict[x]

    if multi_found:
        # Convert non-dicts to dict
        map = {k:(v if (isinstance(v,dict) or pd.isnull(v)) else {0:v}) for k,v in map.items()}

    return col.map(map)


def std_list(col, vals, map_dict, delim, converter, *args, **kwargs):
    map = {}
    multi_found = False
    # Look for multiplication pattern such as F x 2 for 2 females
    re_mult = re.compile(r"([A-Za-z])\s?[Xx]\s?(\d+)")
    re_mult_reverse = re.compile(r"(\d+)\s?[Xx]\s?([A-Za-z])")
    for x in vals:
        if isinstance(x,str):
            xorig = x
            for m in re_mult.finditer(x):
                multi_found = True
                new_str = delim.join([m.group(1) for _ in range(int(m.group(2)))])
                x = x.replace(m.group(0), new_str)
            for m in re_mult_reverse.finditer(x):
                multi_found = True
                new_str = delim.join([m.group(2) for _ in range(int(m.group(1)))])
                x = x.replace(m.group(0), new_str)

            items = x.split(delim)
            new_val = {}
            for k, i in enumerate(items):
                if i == "ISL":  # LA County code for AAPI is ASIAN-PACIFIC,ISL
                    continue
                i = i.strip()
                if i not in map_dict:
                    map_dict[i] = converter(i, *args, **kwargs)
                if len(items)==1:
                    new_val = map_dict[i]
                else:
                    new_val[k] = map_dict[i]
                    multi_found = True

            map[xorig] = new_val
        else:
            if x not in map_dict:
                map_dict[x] = converter(x, *args, **kwargs)
                map[x] = map_dict[x]

    if multi_found:
        # Convert non-dicts to dict
        map = {k:(v if (isinstance(v,dict) or pd.isnull(v)) else {0:v}) for k,v in map.items()}

    return col.map(map)


def std_with_counts(col, vals, map_dict, delim, converter, no_id, *args, **kwargs):
    map = {}
    multi_found = False
    no_id_error = 'error'
    for x in vals:
        if isinstance(x,str):
            x = x.strip()
            try:
                map_dict[x] = converter(x, no_id_error, *args, **kwargs)
                map[x] = map_dict[x]
            except:
                # This needs parsed. Strings can be injuries to multiple delimited by spaces (i.e. 'Deceased Injured')
                # and/or counts of an injury (i.e. '2 Injured')
                items = x.split(delim)
                subject_count = 0
                count = 1
                cur_item = ''
                new_val = {}
                for i in items:
                    if i.isdigit():
                        count = int(i)
                    else:
                        cur_item = cur_item + " " + i.strip() if len(cur_item)>0 else i.strip()
                        try:
                            if cur_item not in map_dict:
                                map_dict[cur_item] = converter(cur_item, no_id_error, *args, **kwargs)
                        except:
                            continue
                        
                        for j in range(0, count):
                            new_val[subject_count+j] = map_dict[cur_item]
                        subject_count+=count
                        count = 1
                        cur_item = ''
                        multi_found = True

                if len(cur_item)>0:
                    if no_id in ['error','test']:
                        raise ValueError("Unable to parse subject injury")
                    elif no_id=='null':
                        new_val = ""
                    else:
                        new_val = x
                map[x] = new_val
        else:
            if x not in map_dict:
                map_dict[x] = converter(x, *args, **kwargs)
                map[x] = map_dict[x]

    if multi_found:
        # Convert non-dicts to dict
        map = {k:(v if (isinstance(v,dict) or pd.isnull(v)) else {0:v}) for k,v in map.items()}

    return col.map(map)


import re

def camel_case_split(x):
    if len(x)==0:
        return x
    words = []
    cur = x[0]
    for l in x[1:]:
        if l.isupper() and cur[-1].islower():
            words.append(cur)
            cur = l
        else:
            cur+=l

    words.append(cur)
    return words


def split_words(string, case=None):
    # Split based on spaces, punctuation, and camel case
    words = list(re.split(r"[^A-Za-z]+", string))
    k = 0
    while k < len(words):
        if len(words[k])==0:
            del words[k]
            continue
        new_words = camel_case_split(words[k])
        words[k] = new_words[0]
        for j in range(1, len(new_words)):
            words.insert(k+1, new_words[j])
            k+=1
        k+=1

    if case!=None:
        if case.lower()=='lower':
            words = [x.lower() for x in words]
        elif case.lower()=='upper':
            words = [x.upper() for x in words]
        else:
            raise ValueError("Unknown input case")

    return words

def is_str_number(x):
    is_num = isinstance(x,str)
    if is_num:
        x = x.strip()
        if not x.isdigit():
            # Try removing decimal
            dec = x.find('.')
            is_num = dec>0 and x[:dec].isdigit() and x[dec+1:].isdigit()
    return is_num


import pytest

if __name__ == "__main__":
	import sys
	sys.path.append('../openpolicedata')
from openpolicedata import data, data_loaders
from openpolicedata import datasets
from openpolicedata.defs import MULTI, DataType
from openpolicedata.exceptions import OPD_DataUnavailableError, OPD_TooManyRequestsError,  \
	OPD_MultipleErrors, OPD_arcgisAuthInfoError, OPD_SocrataHTTPError, OPD_FutureError, OPD_MinVersionError, \
	DateFilterException
import random
from datetime import datetime
import pandas as pd
from time import sleep
import warnings
import os
import re

from test_utils import check_for_dataset

# Set Arcgis data loader to validate queries with arcgis package if installed
data_loaders._verify_arcgis = True

sleep_time = 0.1
log_filename = f"pytest_url_errors_{datetime.now().strftime('%Y%m%d_%H')}.txt"
log_folder = os.path.join(".","data/test_logs")

outages_file = os.path.join("..","opd-data","outages.csv")
# if has_outages:=os.path.exists(outages_file):
has_outages=os.path.exists(outages_file)
if has_outages:
	outages = pd.read_csv(outages_file)
else:
	try:
		outages = pd.read_csv('https://raw.githubusercontent.com/openpolicedata/opd-data/main/outages.csv')
		has_outages = True
	except:
		pass
	
warn_errors = (OPD_DataUnavailableError, OPD_SocrataHTTPError, OPD_FutureError, OPD_MinVersionError)

def test_bloomington_citations():
	if check_for_dataset('Bloomington', 'CITATIONS'):
		src = data.Source("Bloomington")
		count = src.get_count('CITATIONS', MULTI)

		match = src.datasets['TableType']=='CITATIONS'
		assert match.sum()==1
		dataset = src.datasets[match].iloc[0]
		loader = data_loaders.Socrata(dataset['URL'], dataset['dataset_id'])
		df = loader.load()
		assert len(df) == count
		dates = df[dataset['date_field']]

		count_by_year = []
		for y in range(2016, 2021):
			count_by_year.append(src.get_count('CITATIONS', y))
			matches_YYYY = dates.str.startswith(f"{y}-")
			dates = dates.drop(index=matches_YYYY[matches_YYYY].index)
			matches_YY = dates.str.match(r'\d{1,2}/\d{1,2}/'+str(y)[2:])
			dates = dates.drop(index=matches_YY[matches_YY].index)
			assert count_by_year[-1] == matches_YYYY.sum() + matches_YY.sum()

		assert count==sum(count_by_year)

@pytest.mark.slow(reason="This is a slow test that should be run before a major commit.")
def test_load_year(datasets, source, start_idx, skip, loghtml, query={}):
	max_count = 1e5
		
	caught_exceptions = []
	caught_exceptions_warn = []
	already_run = []
	for i in range(len(datasets)):
		if skip != None and datasets.iloc[i]["SourceName"] in skip:
			continue
		if i < start_idx:
			continue
		if source != None and datasets.iloc[i]["SourceName"] != source:
			continue

		if is_stanford(datasets.iloc[i]["URL"]):
			# There are a lot of data sets from Stanford, no need to run them all
			# Just run approximately a small percentage
			rnd = random.uniform(0,1)
			if rnd>0.05:
				continue

		match = True
		for k,v in query.items():
			if datasets.iloc[i][k]!=v:
				match = False
				break
		if not match:
			continue

		srcName = datasets.iloc[i]["SourceName"]
		state = datasets.iloc[i]["State"]

		if datasets.iloc[i]["Agency"] == MULTI and \
			srcName == "Virginia":
			# Reduce size of data load by filtering by agency
			agency = "Henrico Police Department"
		else:
			agency = None

		table_print = datasets.iloc[i]["TableType"]

		unique_id = [srcName, state, datasets.iloc[i]["Agency"], table_print]
		if unique_id in already_run:
			continue

		now = datetime.now().strftime("%d.%b %Y %H:%M:%S")
		print(f"{now} Testing {i} of {len(datasets)-1}: {srcName} {table_print} table")

		src = data.Source(srcName, state=state)

		# Handle cases where URL is required to disambiguate requested dataset
		ds_filter, _ = src._Source__filter_for_source(datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"], None, None, errors=False)
		ds_input = datasets.iloc[[i]] if isinstance(ds_filter,pd.DataFrame) and len(ds_filter)>1 else None

		if ds_input is None:
			already_run.append(unique_id)

		try:
			try:
				years = src.get_years(datasets.iloc[i]["TableType"], datasets=ds_input)
			except ValueError as e:
				if len(e.args)>0 and "Extracting the years" in e.args[0]:
					# Just test reading in the table and continue
					table = src.load(datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"], 
									agency=agency, pbar=False)
					continue
				else:
					raise
			except:
				raise
		except warn_errors as e:
			e.prepend(f"Iteration {i}", srcName, datasets.iloc[i]["TableType"])
			caught_exceptions_warn.append(e)
			continue
		except (OPD_TooManyRequestsError, OPD_arcgisAuthInfoError) as e:
			# Catch exceptions related to URLs not functioning
			e.prepend(f"Iteration {i}", srcName, datasets.iloc[i]["TableType"])
			caught_exceptions.append(e)
			continue
		except:
			raise

		# Adding a pause here to prevent issues with requesting from site too frequently
		sleep(sleep_time)

		years_orig = years.copy()
		for y in src.datasets[src.datasets["TableType"] == datasets.iloc[i]["TableType"]]["Year"]:
			if y != MULTI:
				years.remove(y)

		if len(years)==0:
			# Each datasets is annual
			# Run 1st and last year
			multi_case = False
			years = [min(years_orig)]
			if len(years_orig)>1:
				years.append(max(years_orig))
		else:
			multi_case = True
			if len(years)>1:
				# It is preferred to to not use first or last year that start and stop of year are correct
				years = years[-2:-1]
			else:
				years = years[:1]

		tables = []
		future_error = False
		for year in years:
			try:
				table = src.load(datasets.iloc[i]["TableType"], year, 
										agency=agency, pbar=False, 
										sortby="date",
										nrows=max_count if datasets.iloc[i]["DataType"] not in ["CSV","Excel"] else None)
			except OPD_FutureError as e:
				future_error = True
				break
			except (OPD_DataUnavailableError, OPD_SocrataHTTPError) as e:
				caught_exceptions_warn.append(e)
				tables.append(None)
				continue
			except:
				raise

			sleep(sleep_time)

			if len(table.table)==0:
				# Ensure count should have been 0
				count = src.get_count(datasets.iloc[i]["TableType"], year, agency=agency, force=True)
				if count!=0:
					raise ValueError(f"Expected data for year {year} but received none")
				
				# There may not be any data for the year requested.
				for y in years_orig:
					if y not in years:
						count = src.get_count(datasets.iloc[i]["TableType"], y, agency=agency, force=True)
						if count>0:
							years = [x if x!=year else y for x in years]
							table = src.load(datasets.iloc[i]["TableType"], y, 
										agency=agency, pbar=False, 
										sortby="date",
										nrows=max_count if datasets.iloc[i]["DataType"] not in ["CSV","Excel"] else None)
							break
				else:
					if len(table.table)==0 and has_outages and \
						(outages[["State","SourceName","Agency","TableType","Year"]] == datasets.iloc[i][["State","SourceName","Agency","TableType","Year"]]).all(axis=1).any():
						pass
					else:
						raise ValueError("Unable to find data for any year")

			tables.append(table)

		if future_error:
			continue

		for year in years:
			print(f"Testing for year {year}")

			table = tables[years.index(year)]
			if table is None: # Data could not be loaded
				continue

			if len(table.table)==0 and has_outages and \
				(outages[["State","SourceName","Agency","TableType","Year"]] == datasets.iloc[i][["State","SourceName","Agency","TableType","Year"]]).all(axis=1).any():
				caught_exceptions_warn.append(f'Outage continues for {str(datasets.iloc[i][["State","SourceName","Agency","TableType","Year"]])}')
				continue
			else:
				assert len(table.table)>0

			if table.date_field == None or datasets.iloc[i]["DataType"]==DataType.EXCEL.value or \
				table.date_field.lower()=="year":
				continue

			dts = table.table[table.date_field]
			# Remove all non-datestamps
			dts = dts[dts.apply(lambda x: isinstance(x,pd.Timestamp) or isinstance(x,pd.Period))].convert_dtypes()
			try:
				dts = dts.sort_values(ignore_index=True)
			except TypeError as e:
				if re.search(r"not supported between instances of '(Timestamp|Period)' and '(Timestamp|Period)'", str(e)):
					dts = dts[dts.apply(lambda x: isinstance(x,pd.Timestamp))]
					dts = dts.sort_values(ignore_index=True)
				else:
					raise

			all_years = dts.dt.year.unique().tolist()
			
			try:
				assert len(all_years) == 1
			except AssertionError as e:
				# Some datasets filter by local time but return
				# UTC time
				# Until this is solved more elegantly, removing years between {year}-01-01 00:00:00
				# and {year}-01-01 08:00:00
				dts = dts[(dts < f"{year+1}-01-01 00:00:00") | (dts > f"{year+1}-01-01 08:00:00")]
				all_years = dts.dt.year.unique().tolist()
				assert len(all_years) == 1
			except:
				raise(e)
			assert all_years[0] == year

			if not multi_case:
				continue

			if "month" in table.date_field.lower() or "year" in table.date_field.lower() or "yr" in table.date_field.lower():
				# Cannot currently filter only by month/year
				continue
			
			start_date = str(year-1) + "-12-29"
			stop_date = datetime.strftime(dts.iloc[0], "%Y-%m-%d")

			try:
				table_start = src.load(datasets.iloc[i]["TableType"], [start_date, stop_date], 
												agency=agency, pbar=False)
			except ValueError as e:
				if str(e).startswith('Year range cannot contain the year corresponding to a single year dataset'):
					start_date  = str(year) + "-01-01"
					table_start = src.load(datasets.iloc[i]["TableType"], [start_date, stop_date], 
												agency=agency, pbar=False)
				else:
					raise
			except DateFilterException as e:
				continue

			sleep(sleep_time)
			dts_start = table_start.table[table.date_field]
			dts_start = dts_start[dts_start.apply(lambda x: isinstance(x,pd._libs.tslibs.timestamps.Timestamp))].convert_dtypes()
			dts_start = dts_start.sort_values(ignore_index=True, na_position="first")

			# If this isn't true then the stop date is too early
			assert dts_start.iloc[-1].year == year

			# Find first value date in year
			dts_start = dts_start[dts_start.dt.year == year]
			try:
				assert dts.iloc[0] == dts_start.iloc[0]
			except AssertionError as e:
				# See comments in above try/except
				assert dts.iloc[0].tz_localize(None) <= datetime.strptime(f"{year}-01-01 08:00:00", "%Y-%m-%d %H:%M:%S")
			except:
				raise(e)
			
			if len(table.table) == max_count:
				# Whole dataset was not read. Don't compare to latest data in the year
				continue

			start_date = datetime.strftime(dts.iloc[-1], "%Y-%m-%d")
			stop_date  = str(year+1) + "-01-10"  

			try:
				table_stop = src.load(datasets.iloc[i]["TableType"], [start_date, stop_date], 
												agency=agency, pbar=False)
			except ValueError as e:
				if str(e).startswith('There is more than one source matching') or \
					str(e).startswith('Year range cannot contain the year corresponding to a single year dataset'):
					stop_date  = str(year) + "-12-31"  
					table_stop = src.load(datasets.iloc[i]["TableType"], [start_date, stop_date], 
												agency=agency, pbar=False)
				else:
					raise
			sleep(sleep_time)
			dts_stop = table_stop.table[table.date_field]

			dts_stop = dts_stop[dts_stop.apply(lambda x: not isinstance(x,str))]
			try:
				dts_stop = dts_stop.sort_values(ignore_index=True)
			except TypeError as e:
				dts_stop = dts_stop[dts_stop.apply(lambda x: isinstance(x,pd.Timestamp))]
				dts_stop = dts_stop.sort_values(ignore_index=True)

			# If this isn't true then the start date is too late
			assert dts_stop.iloc[0].year == year

			# Find last value date in year
			dts_stop = dts_stop[dts_stop.dt.year == year]
			assert dts.iloc[-1] == dts_stop.iloc[-1]

	if loghtml:
		log_errors_to_file(caught_exceptions, caught_exceptions_warn)
	else:
		if len(caught_exceptions)==1:
			raise caught_exceptions[0]
		elif len(caught_exceptions)>0:
			msg = f"{len(caught_exceptions)} URL errors encountered:\n"
			for e in caught_exceptions:
				msg += "\t" + e.args[0] + "\n"
			raise OPD_MultipleErrors(msg)

		for e in caught_exceptions_warn:
			warnings.warn(str(e))


def test_source_download_not_limitable(datasets, source, start_idx, skip, query={}):		
	for i in range(len(datasets)):
		if skip != None and datasets.iloc[i]["SourceName"] in skip:
			continue
		if i < start_idx:
			continue
		if source != None and datasets.iloc[i]["SourceName"] != source:
			continue

		if datasets.iloc[i]["DataType"] == DataType.CSV and ".zip" in datasets.iloc[i]["URL"]:
			if is_stanford(datasets.iloc[i]["URL"]):
				# There are a lot of data sets from Stanford, no need to run them all
				# Just run approximately 10%
				rnd = random.uniform(0,1)
				if rnd>0.05:
					continue

			match = True
			for k,v in query.items():
				if datasets.iloc[i][k]!=v:
					match = False
					break
			if not match:
				continue

			srcName = datasets.iloc[i]["SourceName"]
			state = datasets.iloc[i]["State"]
			src = data.Source(srcName, state=state)

			year = datasets.iloc[i]["Year"]
			table_type = datasets.iloc[i]["TableType"]

			# Handle cases where URL is required to disambiguate requested dataset
			ds_filter, _ = src._Source__filter_for_source(datasets.iloc[i]["TableType"], datasets.iloc[i]["Year"], None, None, errors=False)
			url_contains = datasets.iloc[i]['URL'] if isinstance(ds_filter,pd.DataFrame) and len(ds_filter)>1 else None
			id_contains = datasets.iloc[i]['dataset_id'] if isinstance(ds_filter,pd.DataFrame) and len(ds_filter)>1 else None

			now = datetime.now().strftime("%d.%b %Y %H:%M:%S")
			print(f"{now} Testing {i+1} of {len(datasets)}: {srcName}, {state} {table_type} table for {year}")
			try:
				table = src.load(table_type, year, pbar=False, url_contains=url_contains, id_contains=id_contains)
			except OPD_FutureError:
				continue
			except:
				raise

			sleep(sleep_time)

			assert len(table.table)>1
			if not pd.isnull(table.date_field):
				assert table.date_field in table.table
				#assuming a Pandas string dtype('O').name = object is okay too
				assert (table.table[table.date_field].dtype.name in ['datetime64[ns]', 'datetime64[ms]'])
			if not pd.isnull(datasets.iloc[i]["agency_field"]):
				assert datasets.iloc[i]["agency_field"] in table.table


def is_stanford(url):
	return "stanford.edu" in url

def log_errors_to_file(*args):
	if not os.path.exists(log_folder):
		os.mkdir(log_folder)

	filename = os.path.join(log_folder, log_filename)

	if os.path.exists(filename):
		perm = "r+"
	else:
		perm = "w"

	with open(filename, perm) as f:
		for x in args:
			for e in x:
				new_line = ', '.join([str(x) for x in e.args])
				skip = False
				if perm == "r+":
					for line in f:
						if new_line in line or line in new_line:
							skip = True
							break

				if not skip:
					f.write(new_line)
					f.write("\n")

if __name__ == "__main__":
	from test_utils import get_datasets
	# For testing
	use_changed_rows = False
	csvfile = None
	# csvfile = os.path.join('..','opd-data', 'opd_source_table.csv')
	start_idx = 0
	skip = None
	# skip = "Sacramento,Beloit,Rutland"
	source = None
	# source = "Asheville"
	query = {}
	# query = {'DataType':'CSV'}

	datasets = get_datasets(csvfile, use_changed_rows)

	test_load_year(datasets, source, start_idx, skip, False, query=query) 
	start_idx = 0
	test_source_download_not_limitable(datasets, source, start_idx, skip) 



if __name__ == "__main__":
	import sys
	sys.path.append('../openpolicedata')
import pandas as pd
import warnings
import openpolicedata as opd
from openpolicedata.defs import TableType
from openpolicedata.deprecated._decorators import deprecated, input_swap
from openpolicedata.deprecated.datasetsCompat import datasets_query
from openpolicedata.deprecated._pandas import DeprecationHandlerDataFrame, DeprecationHandlerSeries
from openpolicedata.deprecated.source_table_compat import check_compat_source_table
import pytest

@pytest.fixture(scope='module')
def df_compat():
	compat_versions_file = 'https://github.com/openpolicedata/opd-data/raw/main/compatibility/compat_versions.csv'
	return pd.read_csv(compat_versions_file, dtype=str)

@input_swap([0,1], ['table_type','year'], [TableType, {'values':[opd.defs.NA, opd.defs.MULTI], 'types':[list, int]}], opt1=None)
def fswap(table_type,year):
	if table_type:
		TableType(table_type)
	assert year in [opd.defs.NA, opd.defs.MULTI] or isinstance(year,int) or isinstance(year,list)

@input_swap([0,1], ['table_type','year'], [TableType, {'values':[opd.defs.NA, opd.defs.MULTI], 'types':[list, int]}], error=True, opt1=None)
def fswap_error(table_type,year):
	fswap(table_type,year)

@pytest.mark.parametrize("year", [2019, [2019, 2020], opd.defs.NA, opd.defs.MULTI])
@pytest.mark.parametrize("table_type", [TableType.ARRESTS, str(TableType.ARRESTS)])
def test_no_inputswap(year, table_type):
	with warnings.catch_warnings():
		warnings.simplefilter("error")
		fswap(table_type, year)


@pytest.mark.parametrize("year", [2019, [2019, 2020], opd.defs.NA, opd.defs.MULTI])
@pytest.mark.parametrize("table_type", [TableType.ARRESTS, str(TableType.ARRESTS)])
def test_inputswap_warning(year, table_type):
	with pytest.deprecated_call():
		fswap(year, table_type)


@pytest.mark.parametrize("year", [2019, [2019, 2020], opd.defs.NA, opd.defs.MULTI])
@pytest.mark.parametrize("table_type", [TableType.ARRESTS, str(TableType.ARRESTS)])
def test_inputswap_keyword_warning(year, table_type):
	with pytest.deprecated_call():
		fswap(year, table_type=table_type)


@pytest.mark.parametrize("year", [2019, [2019, 2020], opd.defs.NA, opd.defs.MULTI])
@pytest.mark.parametrize("table_type", [TableType.ARRESTS, str(TableType.ARRESTS)])
def test_inputswap_singleinput_warning(year, table_type):
	with pytest.deprecated_call():
		fswap(year)


@pytest.mark.parametrize("year", [2019, [2019, 2020], opd.defs.NA, opd.defs.MULTI])
@pytest.mark.parametrize("table_type", [TableType.ARRESTS, str(TableType.ARRESTS)])
def test_inputswap_error(year, table_type):
	with pytest.raises(ValueError, match='have been swapped'):
		fswap_error(year, table_type)


@deprecated("MSG")
def fdep():
	pass

@pytest.mark.parametrize("func", [fdep, datasets_query])
def test_deprecated_decorator(func):
	with pytest.deprecated_call():
		func()


@pytest.mark.parametrize("type1, type2", [("COMPLAINTS - SUBJECTS", 'COMPLAINTS - CIVILIANS'),
										  ("USE OF FORCE - SUBJECTS/OFFICERS", 'USE OF FORCE - CIVILIANS/OFFICERS')])
def test_deprecated_enums(type1, type2):
	with pytest.deprecated_call():
		assert TableType(type1) == TableType(type2)


def test_datasets_no_civilian_tabletypes():
	df = opd.datasets.query()
	assert not df["TableType"].str.contains("CIVILIAN").any()


def test_datasets_equals_civilian():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	s = df["TableType"]
	assert isinstance(s, DeprecationHandlerSeries)
	with pytest.deprecated_call():
		t = s == "USE OF FORCE - CIVILIANS/OFFICERS"

	assert t.sum()>0
	s = df["TableType"][t]
	assert isinstance(s, pd.Series)
	assert len(s)>0
	assert len(df[t])>0


def test_datasets_equals_subjects():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	s = df["TableType"]
	assert isinstance(s, DeprecationHandlerSeries)
	with warnings.catch_warnings():
		warnings.simplefilter("error")
		t = s == "USE OF FORCE - SUBJECTS/OFFICERS"

	assert t.sum()>0
	assert len(df["TableType"][t])>0
	assert len(df[t])>0


def test_datasets_get_datatype():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	s = df["DataType"]
	assert isinstance(s, pd.Series)


def test_datasets_iloc_single_table_type():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	k = [j for j,x in enumerate(df.columns) if x=="TableType"][0]
	assert isinstance(df.iloc[:, k], DeprecationHandlerSeries)

def test_datasets_iloc_single_table_type_subset_no_subject():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	k = [j for j,x in enumerate(df.columns) if x=="TableType"][0]
	s = df.iloc[:2, k]
	if not s.str.contains("SUBJECT").any():
		assert isinstance(s, pd.Series)

def test_datasets_iloc_single_table_type_subset_subject():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	k = [j for j,x in enumerate(df.columns) if x=="TableType"][0]
	s = df.iloc[31:34, k]
	if s.str.contains("SUBJECT").any():
		assert isinstance(s, DeprecationHandlerSeries)


def test_datasets_iloc_single_not_table_type():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	k = [j for j,x in enumerate(df.columns) if x!="TableType"][0]
	assert isinstance(df.iloc[:, k], pd.Series)

def test_datasets_iloc_single_row():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	assert isinstance(df.iloc[0], pd.Series)


def test_datasets_loc_single_table_type():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	assert isinstance(df.loc[:, "TableType"], DeprecationHandlerSeries)

def test_datasets_loc_single_table_type_subset_no_subject():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	s = df.loc[:2, "TableType"]
	if not s.str.contains("SUBJECT").any():
		assert isinstance(s, pd.Series)

def test_datasets_loc_single_table_type_subset_subject():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	s = df.loc[31:34, "TableType"]
	if s.str.contains("SUBJECT").any():
		assert isinstance(s, DeprecationHandlerSeries)


def test_datasets_loc_single_not_table_type():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	assert isinstance(df.loc[:, "DataType"], pd.Series)

def test_datasets_loc_single_row():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	assert isinstance(df.loc[0], pd.Series)


def test_datasets_isin_civilians():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	with pytest.deprecated_call():
		df["TableType"].isin(["ARRESTS", "OFFICER-INVOLVED SHOOTINGS - CIVILIANS"])


def test_datasets_isin_subjects():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	with warnings.catch_warnings():
		warnings.simplefilter("error")
		df["TableType"].isin(["ARRESTS", "OFFICER-INVOLVED SHOOTINGS - SUBJECTS"])


def test_datasets_isin_DataType():
	df = opd.datasets.query()
	assert isinstance(df, DeprecationHandlerDataFrame)

	with warnings.catch_warnings():
		warnings.simplefilter("error")
		df["DataType"].isin(["ARRESTS", "OFFICER-INVOLVED SHOOTINGS - SUBJECTS"])


def test_pandas_query_no_subject():
	df = opd.datasets.query(state="Virginia")
	assert isinstance(df, pd.DataFrame)


def test_pandas_query_has_subject():
	df = opd.datasets.query(state="California")
	assert isinstance(df, DeprecationHandlerDataFrame)


def test_pandas_query_tabletype_subject():
	with pytest.deprecated_call():
		df = opd.datasets.query(table_type="COMPLAINTS - CIVILIANS")
	assert isinstance(df, DeprecationHandlerDataFrame)
	assert len (df)>0


def test_pandas_query_tabletype_no_subject():
	with warnings.catch_warnings():
		warnings.simplefilter("error")
		df = opd.datasets.query(table_type=opd.defs.TableType.COMPLAINTS_SUBJECTS)
	assert isinstance(df, pd.DataFrame)
	assert len (df)>0

def test_tabletype_contains_subject():
	with pytest.deprecated_call():
		t = opd.datasets.get_table_types(contains="- CIVILIANS")
	assert len (t)>0


def test_tabletype_contains_no_subject():
	with warnings.catch_warnings():
		warnings.simplefilter("error")
		t = opd.datasets.get_table_types(contains="- SUBJECTS")
	assert len (t)>0

def test_source_table_not_deprecated():
	assert not check_compat_source_table(cur_ver='100.0')[0]

def test_source_table_bad_df_compat():
	assert not check_compat_source_table(df_compat=1)[0]

def test_source_table_deprecated():
	with pytest.deprecated_call():
		loaded, df = check_compat_source_table(cur_ver='0.0')
	assert loaded
	assert len(df)==1  # Number of rows in 1st test source table

def test_source_table_fail_not_req(df_compat):
	df_compat = df_compat.copy(deep=True)
	df_compat.loc[0, 'csv_name'] = '?!~notafile.csv'
	with pytest.deprecated_call():
		loaded, df = check_compat_source_table(df_compat=df_compat, cur_ver='0.0')
	assert loaded
	assert len(df)==2  # Number of rows in 2nd test source table

def test_source_table_fail_req(df_compat):
	df_compat = df_compat.copy(deep=True)
	df_compat.loc[1, 'csv_name'] = '?!~notafile.csv'
	with pytest.raises(opd.exceptions.CompatSourceTableLoadError):
		check_compat_source_table(df_compat=df_compat, cur_ver='0.0.1')

@pytest.mark.parametrize("y", [7, 13, range(0,8), [6, 7,13], slice(0,8)])
@pytest.mark.filterwarnings("error::DeprecationWarning")
def test_datasets_iloc_no_warning(all_datasets, y):
	all_datasets.iloc[0,y]


@pytest.mark.parametrize("x", [0, [0,1], range(0,2)])
@pytest.mark.filterwarnings("error::DeprecationWarning")
def test_datasets_iloc_single_input_no_warning(all_datasets, x):
	all_datasets.iloc[x]

@pytest.mark.parametrize("y", [8, 10, [8,9], range(8,12), [1,8], range(6,10), slice(8,10), slice(1,10,2)])
def test_datasets_iloc_warning(all_datasets, y):
	with pytest.deprecated_call():
		all_datasets.iloc[0,y]

if __name__ == "__main__":
	csvfile = None
	csvfile = "C:\\Users\\matth\\repos\\opd-data\\opd_source_table.csv"
	test_pandas_query_tabletype_subject(csvfile,None,None,None,None)
	test_pandas_query_tabletype_no_subject(csvfile,None,None,None,None)




CIV_DEPRECATION_MESSAGE = "The word 'CIVILIAN' has been replaced with 'SUBJECT' in " +\
    f"TableType names to be consistent with the Stanford Open Policing Project "+\
    f"(https://github.com/stanford-policylab/opp/blob/master/data_readme.md). "

CIV_REPLACEMENT_MESSAGE =  "{} is being corrected to {}. "+\
                    "Please replace 'CIVILIAN' with 'SUBJECT' in your code as this correction will be removed "+\
                    "in a future release."


import pytest

import numbers
import pandas as pd

import openpolicedata as opd

@pytest.fixture(scope="module")
def dates():
    src = opd.Source('Richmond')
    t = src.load('OFFICER-INVOLVED SHOOTINGS','MULTIPLE')
    return t.table[t.date_field]

# # data._check_date
# # datetime_parser
# #   parse_date_to_datetime
# #   merge_date_and_time
# #   validate_date
# #   validate_time
# #   parse_time

@pytest.mark.parametrize('format', ['%Y%m%d', '%Y%m%d.0', '%#m%d%Y.0', '%B %#d, %Y', '%#m/%#d/%y', '%#m/%#d/%Y',
                                    '%m-%d-%Y', '%Y-%m-%d', '%m/%d/%Y  00:00', '%Y-%m-%d 00:00:00+00'])
def test_dates_to_datetime(dates, format):
    new_dates = dates.dt.strftime(format)
    dates_conv = opd.datetime_parser.to_datetime(new_dates)
    pd.testing.assert_series_equal(pd.to_datetime(dates.dt.date), dates_conv.dt.tz_localize(None))

@pytest.mark.parametrize('format', ['%Y-%m-%d %H:%M:%S', '%Y-%m-%dT%H:%M:%S.000', '%Y-%m-%dT%H:%M:%S.000Z',
                                    '%m/%d/%Y %H%M hours', '%Y%m%d%H%M%S'])
def test_datetimes_to_datetime(dates, format):
    new_dates = dates.dt.strftime(format)
    dates_conv = opd.datetime_parser.to_datetime(new_dates)
    pd.testing.assert_series_equal(dates, dates_conv.dt.tz_localize(None))

def test_ymd_to_datetime(dates):
    dates_conv = opd.datetime_parser.to_datetime({'year':dates.dt.year,'month':dates.dt.month, 'day':dates.dt.day})
    dates_conv.name = dates.name
    pd.testing.assert_series_equal(pd.to_datetime(dates.dt.date), dates_conv)

@pytest.mark.parametrize('ordinal',[False, True])
@pytest.mark.parametrize('object',[False, True])
def test_ymd_ordinal_to_datetime(dates, object, ordinal):
    ymd = {'year':dates.dt.year.copy(),'month':dates.dt.month.copy(), 'day':dates.dt.day.copy()}
    dates_only = pd.to_datetime(dates.dt.date).astype('object')
    if ordinal:
        ymd['day'] = ymd['day'].astype('object')
        ords = ['st','nd','rd','th','th']  # Added extra th so number of ords and null cases are not equal
        for k in range(len(dates)):
            if pd.notnull(ymd['day'][k]):
                ymd['day'][k] = str(ymd['day'][k]) + ords[k%len(ords)]
                
    for k in range(len(dates)):
        for m in range(k%4):
            ymd[list(ymd.keys())[2-m]][k] = pd.NA

        if k%4==1:
            dates_only[k] = pd.Period(dates_only[k],'M')
        elif k%4==2:
            dates_only[k] = pd.Period(dates_only[k],'Y')
        elif k%4==3:
            dates_only[k] = pd.NaT

    df = pd.DataFrame(ymd)
    if object:
        df['day'] = df['day'].astype('object')
        df.loc[df['day'].isnull(),'day'] = pd.NA
    dates_conv = opd.datetime_parser.to_datetime(df)
    dates_conv.name = dates.name
    pd.testing.assert_series_equal(dates_only, dates_conv)

def test_unix_to_datetime(dates):
    unix = dates.apply(lambda x: x.timestamp())*1000
    dates_conv = opd.datetime_parser.to_datetime(unix, unit='ms')
    pd.testing.assert_series_equal(dates, dates_conv)

@pytest.mark.parametrize('format, period', [('%Y-%m-__', 'M'),('%Y-__-__', 'Y'),('%Y-__-%d', 'Y')]) 
def test_blank_to_datetime(dates, format, period):
    periods = dates.dt.to_period(freq=period)
    new_dates = dates.dt.strftime(format)
    dates_conv = opd.datetime_parser.to_datetime(new_dates, ignore_errors=True)
    pd.testing.assert_series_equal(periods, dates_conv)

@pytest.mark.parametrize('format',['%Y-%m-%d %H:%M:%S','%Y-%m-%dT%H:%M:%S.000Z'])
def test_mixed_to_datetime(dates, format):
    dates = dates.copy()

    new_dates = pd.Series({k:v.strftime('%Y-%m-%d') if k/2%1!=0 else v.strftime(format) for k,v in dates.items()}, name=dates.name)
    
    idx = dates.index/2 % 1!=0
    dates[idx] = dates[idx].dt.date
    
    dates_conv = opd.datetime_parser.to_datetime(new_dates)
    pd.testing.assert_series_equal(dates, dates_conv)

def test_mixed_floats_to_datetime(dates):
    new_dates = pd.Series({k:v.strftime('%Y%m%d.0') if k/2%1!=0 else v.strftime('%#m%d%Y.0') for k,v in dates.items()}, name=dates.name)
    dates_conv = opd.datetime_parser.to_datetime(new_dates)
    pd.testing.assert_series_equal(pd.to_datetime(dates.dt.date), dates_conv.dt.tz_localize(None))


# Definition of constants
from __future__ import annotations
from copy import deepcopy
from enum import Enum
import inspect
import pandas as pd
import re
from sys import version_info
import warnings

# These are the types of data currently available in opd.
# They all have corresponding data loaders in data_loaders.py
# When new data loaders are added, this list should be updated.
class DataType(str, Enum):
    EXCEL = "Excel"
    CSV = "CSV"
    ArcGIS = "ArcGIS"
    SOCRATA = "Socrata"
    CARTO = "Carto"
    CKAN = 'CKAN'

# These are the types of tables currently available in opd.
# Add to this list when datasets do not correspond to the below data types
class TableType(str, Enum):
    # Adding a description property in addition to value
    # https://rednafi.github.io/reflections/add-additional-attributes-to-enum-members-in-python.html
    description = None

    def __new__(
        cls, value: str, description: str = ""
    ) -> TableType:

        obj = str.__new__(cls, value)
        obj._value_ = value
        obj.description = description
        return obj
    
    def __str__(self):
        return self.value
    
    
    @classmethod
    def _missing_(cls, value):
        # https://stackoverflow.com/questions/37444002/overriding-enum-call-method
        # Handles deprecation of CIVILIANS usage to SUBJECTS
        if "- CIVILIANS" in value:
            new_value = value.replace("- CIVILIANS", "- SUBJECTS")
            if new_value in TableType._value2member_map_:
                warnings.warn(
                    f"TableType {value} is deprecated. CIVILIAN has been replaced with SUBJECT in TableType names. Requested TableType has been automatically updated to {new_value}.",
                    DeprecationWarning,
                )
                return TableType(new_value)
        
        return super()._missing_(value)

    # Below tuples are (value, description)
    ARRESTS = ("ARRESTS", "Seizures or forcible restraints by police")
    CALLS_FOR_SERVICE = ("CALLS FOR SERVICE", "Includes dispatched calls (911 or non-emergency #) and officer-initiated calls")
    CITATIONS = ("CITATIONS","Commonly referred to as tickets, citations indicate a violation of the law and may be given for "+
        "violations such as traffic and civil violations")
    COMPLAINTS = ("COMPLAINTS","Complaints of police misconduct made internally or by the community")
    COMPLAINTS_ALLEGATIONS = ("COMPLAINTS - ALLEGATIONS",
        "Complaint data may be split into several tables. This table contains specific data on the allegations.")
    COMPLAINTS_BACKGROUND = ("COMPLAINTS - BACKGROUND",
        "Complaint data may be split into several tables. This table contains data on the general background of the complaints.")
    COMPLAINTS_BODY_WORN_CAMERA = ("COMPLAINTS - BODY WORN CAMERA",
        "Complaint data may be split into several tables. This table contains data on body worn camera details.")
    COMPLAINTS_SUBJECTS = ("COMPLAINTS - SUBJECTS",
        "Complaint data may be split into several tables. This table contains specific data on the involved subjects.")
    COMPLAINTS_SUBJECTS_OFFICERS = ("COMPLAINTS - SUBJECTS/OFFICERS",
        "Complaint data may be split into several tables. This table contains data on the involved subjects and officers. "+
        "A row in the data will indicate whether that row corresponds to an officer or subject.")
    COMPLAINTS_OFFICERS = ("COMPLAINTS - OFFICERS",
        "Complaint data may be split into several tables. This table contains specific data on the involved officers.")
    COMPLAINTS_PENALTIES = ("COMPLAINTS - PENALTIES",
        "Complaint data may be split into several tables. This table contains specific data on any resulting penalties.")
    CRASHES = ("CRASHES", "Traffic crashes")
    CRASHES_SUBJECTS = ("CRASHES - SUBJECTS",
        "Crash data may be split into several tables due to the possibility that multiple "+
        "subjects (drivers and/or passengers) and vehicles may be involved in an incident. This table contains data on subjects.")
    CRASHES_INCIDENTS = ("CRASHES - INCIDENTS",
        "Crash data may be split into several tables due to the possibility that multiple "+
        "subjects and vehicles may be involved in an incident. This table contains data on the incident.")
    CRASH_NONMOTORIST = ("CRASHES - NONMOTORIST",
        "Crash data may be split into several tables due to the possibility that multiple "+
        "subjects and vehicles may be involved in an incident. This table contains information on non-motorists involved in a crash.")
    CRASHES_VEHICLES = ("CRASHES - VEHICLES",
        "Crash data may be split into several tables due to the possibility that multiple "+
        "subjects and vehicles may be involved in an incident. This table contains data on vehicles.")
    DISCIPLINARY_RECORDS = ("DISCIPLINARY RECORDS",
        "Disciplinary records of officers")
    EMPLOYEE = ("EMPLOYEE","Demographic data of the police workforce")
    DEATHS_IN_CUSTODY = ("DEATHS IN CUSTODY", "Deaths that occur in custody or during the process of arrest")
    FIELD_CONTACTS = ("FIELD CONTACTS", "Consensual contacts between officers and the community.")
    INCIDENTS = ("INCIDENTS", "Crime incident reports")
    INCIDENTS_INCIDENTS = ("INCIDENTS - INCIDENTS", 
                          "Incidents data may be split into several tables due to the possibility that multiple "+
                        "subjects may be involved in an incident. This table contains data on the details of the incident.")
    INCIDENTS_SUBJECTS = ("INCIDENTS - SUBJECTS", 
                          "Incidents data may be split into several tables due to the possibility that multiple "+
                        "subjects may be involved in an incident. This table contains data on subjects.")
    LAWSUITS = ("LAWSUITS", "Lawsuits against a police department")
    PEDESTRIAN = ("PEDESTRIAN STOPS","Stops of pedestrians based on 'reasonable suspicion'. May lead to a frisk.")
    PEDESTRIAN_ARRESTS = ("PEDESTRIAN ARRESTS","Pedestrian stops leading to an arrest")
    PEDESTRIAN_CITATIONS = ("PEDESTRIAN CITATIONS","Pedestrian stops leading to a citation")
    PEDESTRIAN_WARNINGS = ("PEDESTRIAN WARNINGS","Pedestrian stops leading to a warning")
    POINTING_WEAPON = ('POINTING WEAPON', "Instances of officers pointing a weapon (firearm, Taser, etc.) at individuals.")
    SEARCHES = ("SEARCHES", "Details of searches")
    SHOOTINGS = ("OFFICER-INVOLVED SHOOTINGS","Shootings by officers")
    SHOOTINGS_SUBJECTS = ("OFFICER-INVOLVED SHOOTINGS - SUBJECTS",
        "Officer-involved shootings data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on subjects.")
    SHOOTINGS_OFFICERS = ("OFFICER-INVOLVED SHOOTINGS - OFFICERS",
        "Officer-involved shootings data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on officer.")
    SHOOTINGS_INCIDENTS = ("OFFICER-INVOLVED SHOOTINGS - INCIDENTS",
        "Officer-involved shootings data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on the incident.")
    STOPS = ("STOPS","Contains data on both pedestrian and traffic stops.")
    STOPS_INCIDENTS = ("STOPS - INCIDENTS",
        "Stops data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on the incident.")
    STOPS_SUBJECTS = ("STOPS - SUBJECTS",
        "Stops data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on the subjects.")
    TRAFFIC = ("TRAFFIC STOPS","Traffic stops are stops by police of motor vehicles due to reasonable suspicion " + 
        " or traffic violations.")
    TRAFFIC_INCIDENTS = ("TRAFFIC STOPS - INCIDENTS",
        "Traffic stops data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on the incident.")
    TRAFFIC_SUBJECTS = ("TRAFFIC STOPS - SUBJECTS",
        "Traffic stops data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on the subjects.")
    TRAFFIC_ARRESTS = ("TRAFFIC ARRESTS","Traffic stops leading to an arrest.")
    TRAFFIC_CITATIONS = ("TRAFFIC CITATIONS","Traffic stops leading to a citation.")
    TRAFFIC_WARNINGS = ("TRAFFIC WARNINGS","Traffic stops leading to a warning.")
    USE_OF_FORCE = ("USE OF FORCE","Documentation of physical force used against subjects.")
    USE_OF_FORCE_SUBJECTS = ("USE OF FORCE - SUBJECTS",
        "Use of force data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on subjects.")
    USE_OF_FORCE_OFFICERS = ("USE OF FORCE - OFFICERS",
        "Use of force data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on officer.")
    USE_OF_FORCE_INCIDENTS = ("USE OF FORCE - INCIDENTS",
        "Use of force data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on the incident.")
    USE_OF_FORCE_SUBJECTS_OFFICERS = ("USE OF FORCE - SUBJECTS/OFFICERS",
        "Use of force data may be split into several tables due to the possibility that multiple "+
        "subjects and officers may be involved in an incident. This table contains data on subjects and officers."+
        "A row in the data will indicate whether that row corresponds to an officer or subject.")
    VEHICLE_PURSUITS = ("VEHICLE PURSUITS","Attempts by officers in vehicles to pursue vehicles where the operator " + 
        "is believed to be aware that they are being signaled to stop but who is fleeing or ignoring the officer's attempt "+
        "to stop them.")
    WARNINGS = ("WARNINGS","Warnings for traffic and civil violations")

# Constants used in dataset parameters
MULTI = "MULTIPLE"    # For data sets that put multiple years or agencies in 1 dataset
NA = "NONE"         # None = not applicable (pandas converts "N/A" to NaN)
UNSPECIFIED = "UNSPECIFIED"

states = {
    "Alabama": "AL",
    "Alaska": "AK",
    "Arizona": "AZ",
    "Arkansas": "AR",
    "California": "CA",
    "Colorado": "CO",
    "Connecticut": "CT",
    "Delaware": "DE",
    "Florida": "FL",
    "Georgia": "GA",
    "Hawaii": "HI",
    "Idaho": "ID",
    "Illinois": "IL",
    "Indiana": "IN",
    "Iowa": "IA",
    "Kansas": "KS",
    "Kentucky": "KY",
    "Louisiana": "LA",
    "Maine": "ME",
    "Maryland": "MD",
    "Massachusetts": "MA",
    "Michigan": "MI",
    "Minnesota": "MN",
    "Mississippi": "MS",
    "Missouri": "MO",
    "Montana": "MT",
    "Nebraska": "NE",
    "Nevada": "NV",
    "New Hampshire": "NH",
    "New Jersey": "NJ",
    "New Mexico": "NM",
    "New York": "NY",
    "North Carolina": "NC",
    "North Dakota": "ND",
    "Ohio": "OH",
    "Oklahoma": "OK",
    "Oregon": "OR",
    "Pennsylvania": "PA",
    "Rhode Island": "RI",
    "South Carolina": "SC",
    "South Dakota": "SD",
    "Tennessee": "TN",
    "Texas": "TX",
    "Utah": "UT",
    "Vermont": "VT",
    "Virginia": "VA",
    "Washington": "WA",
    "West Virginia": "WV",
    "Wisconsin": "WI",
    "Wyoming": "WY",
    "District of Columbia": "DC",
    "American Samoa": "AS",
    "Guam": "GU",
    "Northern Mariana Islands": "MP",
    "Puerto Rico": "PR",
    "United States Minor Outlying Islands": "UM",
    "U.S. Virgin Islands": "VI",
}

class _ToDict_Mixin:
    def to_dict(self):
        attributes = inspect.getmembers(self, lambda a:not(inspect.isroutine(a)))
        return {a:v for a,v in attributes if not(a.startswith('__'))}

# Standard column names
class _Columns(_ToDict_Mixin):
    DATE = "DATE"
    TIME = "TIME"
    DATETIME = "DATETIME"
    RACE_SUBJECT = "SUBJECT_RACE"
    RACE_ETHNICITY_SUBJECT = "SUBJECT_RACE/ETHNICITY"
    ETHNICITY_SUBJECT = "SUBJECT_ETHNICITY"
    RACE_OFFICER = "OFFICER_RACE"
    RACE_ETHNICITY_OFFICER = "OFFICER_RACE/ETHNICITY"
    ETHNICITY_OFFICER = "OFFICER_ETHNICITY"
    ETHNICITY_OFFICER_SUBJECT = "OFFICER/SUBJECT_ETHNICITY"
    RACE_OFFICER_SUBJECT = "OFFICER/SUBJECT_RACE"
    RACE_ETHNICITY_OFFICER_SUBJECT = "OFFICER/SUBJECT_RACE/ETHNICITY"
    AGENCY = "AGENCY"
    AGE_SUBJECT = "SUBJECT_AGE"
    AGE_OFFICER = "OFFICER_AGE"
    AGE_OFFICER_SUBJECT = "OFFICER/SUBJECT_AGE"
    AGE_RANGE_SUBJECT = "SUBJECT_AGE_RANGE"
    AGE_RANGE_OFFICER = "OFFICER_AGE_RANGE"
    AGE_RANGE_OFFICER_SUBJECT = "OFFICER/SUBJECT_AGE_RANGE"
    GENDER_SUBJECT = "SUBJECT_GENDER"
    GENDER_OFFICER = "OFFICER_GENDER"
    GENDER_OFFICER_SUBJECT = "OFFICER/SUBJECT_GENDER"
    RE_GROUP_OFFICER = "OFFICER_RE_GROUP"
    RE_GROUP_SUBJECT = "SUBJECT_RE_GROUP"
    RE_GROUP_OFFICER_SUBJECT = "OFFICER/SUBJECT_RE_GROUP"
    SUBJECT_OR_OFFICER = "SUBJECT_OR_OFFICER"
    FATAL_SUBJECT = "SUBJECT_FATAL"
    FATAL_OFFICER = "OFFICER_FATAL"
    FATAL_OFFICER_SUBJECT = "OFFICER/SUBJECT_FATAL"
    INCIDENT_ID = "INCIDENT_ID"
    INJURY_SUBJECT = "SUBJECT_INJURY"
    INJURY_OFFICER = "OFFICER_INJURY"
    INJURY_OFFICER_SUBJECT = "OFFICER/SUBJECT_INJURY"
    NAME_SUBJECT = "SUBJECT_NAME"
    NAME_OFFICER = "OFFICER_NAME"
    NAME_OFFICER_SUBJECT = "OFFICER/SUBJECT_NAME"
    ZIP_CODE = 'ZIP_CODE'

    def _get_columns_as_df(self):
        attributes = self.to_dict()
        props = []
        columns = []
        defs = []
        sort_by = []
        for a, v in attributes.items():
            props.append(a)
            columns.append(v)

            match = re.match(r"^(SUBJECT|OFFICER|OFFICER/SUBJECT)\s(.+)$", v.replace("_"," "))
            if v=='SUBJECT_OR_OFFICER':
                sort_by.append(v)
                defs.append("Whether row describes an officer or an subject/civilian")
            elif match and match.group(2)=="FATAL":
                defs.append(f"Whether {match.group(1).lower()} was fatally injured in an officer-involved shooting or use of force")
                sort_by.append(match.group(2).title())
            elif match and match.group(2)=="INJURY":
                defs.append(f"Whether {match.group(1).lower()} was injured in an officer-involved shooting or use of force")
                sort_by.append(match.group(2).title())
            elif match and match.group(2)=="NAME":
                defs.append(f"Name of {match.group(1).lower()} in an officer-involved shooting")
                sort_by.append(match.group(2).title())
            elif match:
                addon = ''
                if match.group(2)=="RE GROUP":
                    desc = "Convenience column identical to Race/Ethnicity (if there is a merged Race/Ethncity column) or Race (otherwise)"
                else:
                    desc = match.group(2).title()
                if match.group(1)=="OFFICER/SUBJECT":
                    defs.append(f'{desc} of either an officer or subject (depending on column "{self.SUBJECT_OR_OFFICER}")'+addon)
                else:
                    defs.append(f"{desc} of {match.group(1).lower()}"+addon)
                sort_by.append(match.group(2).title())
            else:
                if a=="DATETIME":
                    defs.append("Combination of date and time when both columns are found (not generated when detected date column contains datetime values)")
                elif a=='INCIDENT_ID':
                    defs.append("A unique incident ID given to an incident (arrest, use of force, etc.). "+
                                "It can be used to relate information across tables. Only standardized by Table.merge function.")
                else:
                    defs.append(v.replace("_"," ").title())
                sort_by.append(v.title())

        df = pd.DataFrame({"Attribute":props, "Column Name":columns, "Definition":defs, 'sort_by':sort_by})
        return df.sort_values(by='sort_by').drop(columns='sort_by').reset_index(drop=True)

    def __repr__(self, ) -> str:
        df = self._get_columns_as_df()
        repr_params = pd.io.formats.format.get_dataframe_repr_params()
        return df.to_string(**repr_params)
    
    def _repr_html_(self):
        df = self._get_columns_as_df()
        return df.to_html()
    

columns = _Columns()

class _Races(_ToDict_Mixin):
    AAPI = "AAPI"
    ASIAN = "ASIAN"
    BLACK = "BLACK"
    INDIGENOUS = "INDIGENOUS"
    LATINO = "LATINO"
    MIDDLE_EASTERN = "MIDDLE_EASTERN"
    MIDDLE_EASTERN_SOUTH_ASIAN = "MIDDLE_EASTERN_SOUTH_ASIAN"
    MULTIPLE = "MULTIPLE"
    OTHER = "OTHER"
    OTHER_UNKNOWN = "OTHER / UNKNOWN"
    PACIFIC_ISLANDER = "PACIFIC_ISLANDER"
    SOUTH_ASIAN = "SOUTH_ASIAN"
    UNKNOWN = "UNKNOWN"
    UNSPECIFIED = "UNSPECIFIED"
    WHITE = "WHITE"

_race_keys = _Races()

class _Ethnicities(_ToDict_Mixin):
    LATINO = _race_keys.LATINO
    MIDDLE_EASTERN = _race_keys.MIDDLE_EASTERN
    NONLATINO = "NON-LATINO"
    UNKNOWN = _race_keys.UNKNOWN
    UNSPECIFIED = _race_keys.UNSPECIFIED

_eth_keys = _Ethnicities()

class _Label:
    def __init__(self, full, compact=None) -> None:
        self.full = full
        self.compact = compact if compact else full[0]

_eth_cats_basic = {
    _eth_keys.LATINO:_Label("HISPANIC/LATINO"),
    _eth_keys.MIDDLE_EASTERN:_Label("MIDDLE EASTERN",'ME'),
    _eth_keys.NONLATINO:_Label("NON-HISPANIC/NON-LATINO",'NH'),
    _eth_keys.UNKNOWN:_Label("UNKNOWN","UNKNOWN"),
    _eth_keys.UNSPECIFIED:_Label(UNSPECIFIED,UNSPECIFIED),
}

_race_cats_basic = {
    _race_keys.AAPI:_Label("ASIAN/PACIFIC ISLANDER",'AAPI'),
    _race_keys.ASIAN:_Label("ASIAN"),
    _race_keys.BLACK:_Label("BLACK"),
    _race_keys.LATINO:_Label("HISPANIC/LATINO"),
    _race_keys.MULTIPLE:_Label("MULTIPLE","MULTIPLE"),
    _race_keys.INDIGENOUS:_Label("INDIGENOUS"),
    _race_keys.OTHER:_Label("OTHER"),
    _race_keys.OTHER_UNKNOWN:_Label("OTHER OR UNKNOWN",'OTHER OR UNKNOWN'),
    _race_keys.UNKNOWN:_Label("UNKNOWN","UNKNOWN"),
    _race_keys.UNSPECIFIED:_Label(UNSPECIFIED, UNSPECIFIED),
    _race_keys.WHITE:_Label("WHITE","W")
}

_more_race_cats = {
    _race_keys.PACIFIC_ISLANDER : _Label("HAWAIIAN/PACIFIC ISLANDER","H/PI"),
    _race_keys.MIDDLE_EASTERN:_Label("MIDDLE EASTERN","ME"),
    # Use in CA stops data: https://data-openjustice.doj.ca.gov/sites/default/files/dataset/2023-01/RIPA%20Dataset%20Read%20Me%202021%20Final%20rev%20011223.pdf
    _race_keys.MIDDLE_EASTERN_SOUTH_ASIAN:_Label("MIDDLE EASTERN/SOUTH ASIAN","ME/SA"),
    _race_keys.SOUTH_ASIAN:_Label("SOUTH ASIAN","SA"),
}

# Combine to form _agg_race_cats
if version_info.minor >= 9:
    _race_cats_expanded = _race_cats_basic | _more_race_cats
else:
    _race_cats_expanded = {**_race_cats_basic, **_more_race_cats}

class _Genders(_ToDict_Mixin):
    MALE = "MALE"
    FEMALE = "FEMALE"
    TRANSGENDER_MALE = "TRANSGENDER_MALE"
    TRANSGENDER_FEMALE = "TRANSGENDER_FEMALE"
    TRANSGENDER = "TRANSGENDER"
    GENDER_NONCONFORMING = "GENDER_NONCONFORMING"
    TRANSGENDER_OR_GENDER_NONCONFORMING = "TRANSGENDER_OR_GENDER_NONCONFORMING"
    GENDER_NONBINARY = "GENDER_NONBINARY"
    OTHER = "OTHER"
    UNKNOWN = "UNKNOWN"
    UNSPECIFIED = "UNSPECIFIED"
    UNSPECIFIED_OR_ANOTHER = "UNSPECIFIED_OR_ANOTHER_GENDER_IDENTITY" # https://www.state.gov/x-gender-marker-available-on-u-s-passports-starting-april-11/

_gender_keys = _Genders()

_genders = {
    _gender_keys.MALE:_Label("MALE"),
    _gender_keys.FEMALE:_Label("FEMALE"),
    _gender_keys.TRANSGENDER_MALE:_Label("TRANSGENDER MALE","TM"),
    _gender_keys.TRANSGENDER_FEMALE:_Label("TRANSGENDER FEMALE","TF"),
    _gender_keys.TRANSGENDER:_Label("TRANSGENDER","T"),
    _gender_keys.GENDER_NONCONFORMING:_Label("GENDER NON-CONFORMING","GNC"),
    _gender_keys.TRANSGENDER_OR_GENDER_NONCONFORMING:_Label("TRANSGENDER OR GENDER NON-CONFORMING","T/GNC"),
    _gender_keys.GENDER_NONBINARY:_Label("GENDER NON-BINARY","GNB"),
    _gender_keys.OTHER:_Label("OTHER"),
    _gender_keys.UNKNOWN:_Label("UNKNOWN","UNKNOWN"),
    _gender_keys.UNSPECIFIED:_Label(UNSPECIFIED,UNSPECIFIED),
    _gender_keys.UNSPECIFIED_OR_ANOTHER:_Label('UNSPECIFIED_OR_ANOTHER_GENDER_IDENTITY','X'),
}

class _PersonTypes(_ToDict_Mixin):
    OFFICER = "OFFICER"
    SUBJECT = "SUBJECT"
    UNSPECIFIED = "UNSPECIFIED"

_roles = _PersonTypes()

def get_roles():
    return deepcopy(_roles)

def get_race_keys():
    return deepcopy(_race_keys)

def get_eth_keys():
    return deepcopy(_eth_keys)

def get_race_cats(expand=False, compact=False):
    cats = _race_cats_expanded if expand else _race_cats_basic
    cats = {k: (v.compact if compact else v.full) for k,v in cats.items()}
    assert len(cats) == len(set(cats.values()))
    return cats

def get_eth_cats(compact=False):
    cats = {k: (v.compact if compact else v.full) for k,v in _eth_cats_basic.items()}
    assert len(cats) == len(set(cats.values()))
    return cats

def get_gender_keys():
    return deepcopy(_gender_keys)

def get_gender_cats(compact=False):
    cats = {k: (v.compact if compact else v.full) for k,v in _genders.items()}
    assert len(cats) == len(set(cats.values()))
    return cats


if __name__ == "__main__":
	import sys
	sys.path.append('../openpolicedata')
from openpolicedata import data
from openpolicedata import datasets
from openpolicedata.defs import MULTI, DataType
import openpolicedata as opd
from openpolicedata.exceptions import OPD_DataUnavailableError, OPD_TooManyRequestsError,  \
	OPD_MultipleErrors, OPD_arcgisAuthInfoError, OPD_SocrataHTTPError, OPD_FutureError, OPD_MinVersionError
from datetime import datetime
from time import sleep
import warnings
import os
import pandas as pd
import pytest

from test_utils import check_for_dataset

sleep_time = 0.1
log_filename = f"pytest_url_errors_{datetime.now().strftime('%Y%m%d_%H')}.txt"
log_folder = os.path.join(".","data/test_logs")

# Set Arcgis data loader to validate queries with arcgis package if installed
opd.data_loaders._verify_arcgis = True

outages_file = os.path.join("..","opd-data","outages.csv")
# if has_outages:=os.path.exists(outages_file):
has_outages=os.path.exists(outages_file)
if has_outages:
	outages = pd.read_csv(outages_file)
else:
	try:
		outages = pd.read_csv('https://raw.githubusercontent.com/openpolicedata/opd-data/main/outages.csv')
		has_outages = True
	except:
		pass
	
warn_errors = (OPD_DataUnavailableError, OPD_SocrataHTTPError, OPD_FutureError, OPD_MinVersionError)

multi_states = ['Connecticut','New York','Virginia']
multi_tables = ['TRAFFIC STOPS','TRAFFIC CITATIONS','STOPS']
multi_years = [MULTI,MULTI,MULTI]
multi_partial = ["Hartford",'Buffalo',"Arlington"]

def test_get_years(datasets, source, start_idx, skip, loghtml, query={}):
	caught_exceptions = []
	caught_exceptions_warn = []

	already_ran = []
	last_source = None
	for i in range(len(datasets)):
		if source != None and datasets.iloc[i]["SourceName"] != source:
			continue
		if skip != None and datasets.iloc[i]["SourceName"] in skip:
			continue
		if i < start_idx:
			continue
		if is_filterable(datasets.iloc[i]["DataType"]) or datasets.iloc[i]["Year"] != MULTI or \
			datasets.iloc[i]["DataType"] == DataType.EXCEL.value:  # If Excel, we can possibly check
			srcName = datasets.iloc[i]["SourceName"]
			state = datasets.iloc[i]["State"]

			match = True
			for k,v in query.items():
				if datasets.iloc[i][k]!=v:
					match = False
					break
			if not match:
				continue

			if (srcName, state, datasets.iloc[i]["TableType"]) in already_ran:
				continue

			src = data.Source(srcName, state=state)

			table_print = datasets.iloc[i]["TableType"]
			now = datetime.now().strftime("%d.%b %Y %H:%M:%S")
			print(f"{now} Testing {i+1} of {len(datasets)}: {srcName} {table_print} table")

			if datasets.iloc[i]["DataType"] == DataType.EXCEL.value:
				if pd.notnull(datasets.iloc[i]['dataset_id']) and ';' in datasets.iloc[i]['dataset_id']:
					# Multi-dataset table
					if pd.isnull(datasets.iloc[i]['date_field']):
						continue
				else:
					loader = opd.data_loaders.Excel(datasets.iloc[i]["URL"])
					has_year_sheets = loader._Excel__get_sheets()[1]
					if not has_year_sheets:
						continue				

			already_ran.append((srcName, state, datasets.iloc[i]["TableType"]))

			if srcName == last_source:
				sleep(num_sources*0.1) # Sleep for a bit to not hit the same site repeatedly too hard
				num_sources +=1
			else:
				num_sources = 1
			last_source = srcName

			try:
				years = src.get_years(datasets.iloc[i]["TableType"], force=True)
			except warn_errors as e:
				e.prepend(f"Iteration {i}", srcName, datasets.iloc[i]["TableType"])
				caught_exceptions_warn.append(e)
				continue
			except (OPD_TooManyRequestsError, OPD_arcgisAuthInfoError) as e:
				# Catch exceptions related to URLs not functioning
				e.prepend(f"Iteration {i}", srcName, datasets.iloc[i]["TableType"])
				caught_exceptions.append(e)
				continue
			except:
				raise

			if len(years)==0 and has_outages and \
				(outages[["State","SourceName","Agency","TableType","Year"]] == datasets.iloc[i][["State","SourceName","Agency","TableType","Year"]]).all(axis=1).any():
				caught_exceptions_warn.append(f'Outage continues for {str(datasets.iloc[i][["State","SourceName","Agency","TableType","Year"]])}')
				continue

			if datasets.iloc[i]["Year"] != MULTI:
				assert datasets.iloc[i]["Year"] in years
			else:
				assert len(years) > 0

			# Adding a pause here to prevent issues with requesting from site too frequently
			sleep(0.1)

	if loghtml:
		log_errors_to_file(caught_exceptions, caught_exceptions_warn)
	else:
		if len(caught_exceptions)==1:
			raise caught_exceptions[0]
		elif len(caught_exceptions)>0:
			msg = f"{len(caught_exceptions)} URL errors encountered:\n"
			for e in caught_exceptions:
				msg += "\t" + e.args[0] + "\n"
			raise OPD_MultipleErrors(msg)

		for e in caught_exceptions_warn:
			warnings.warn(str(e))


def test_get_agencies(datasets, source, start_idx, skip):
		
	for i in range(len(datasets)):
		if skip != None and datasets.iloc[i]["SourceName"] in skip:
			continue
		if i < start_idx:
			continue
		if source != None and datasets.iloc[i]["SourceName"] != source:
			continue

		if is_filterable(datasets.iloc[i]["DataType"]) or datasets.iloc[i]["Agency"] != MULTI:
			srcName = datasets.iloc[i]["SourceName"]
			state = datasets.iloc[i]["State"]
			src = data.Source(srcName, state=state)

			table_print = datasets.iloc[i]["TableType"]
			now = datetime.now().strftime("%d.%b %Y %H:%M:%S")
			print(f"{now} Testing {i+1} of {len(datasets)}: {srcName} {table_print} table")

			try:
				agencies = src.get_agencies(datasets.iloc[i]["TableType"], year=datasets.iloc[i]["Year"])
			except ValueError as e:
				if 'Inputs must filter for a single source' in str(e):
					agencies = src.get_agencies(datasets.iloc[i]["TableType"], year=datasets.iloc[i]["Year"], 
								 url_contains=datasets.iloc[i]["URL"], id_contains=datasets.iloc[i]["dataset_id"])
				else:
					raise
			except OPD_MinVersionError: continue
			except: raise

			if datasets.iloc[i]["Agency"] != MULTI:
				assert [datasets.iloc[i]["Agency"]] == agencies
			else:
				assert len(agencies) > 0

			# Adding a pause here to prevent issues with requesting from site too frequently
			sleep(sleep_time)


def test_multi_agency_list(datasets):
	for i in range(len(datasets)):
		if is_filterable(datasets.iloc[i]["DataType"]) and datasets.iloc[i]["Agency"] == MULTI:
			assert any([datasets.iloc[i]['State']==x and datasets.iloc[i]['TableType']==y and datasets.iloc[i]['Year']==z for 
						x,y,z in zip(multi_states, multi_tables,multi_years)])


@pytest.mark.parametrize('state,table_type,year,partial',[(x,y,z,a) for x,y,z,a in zip(multi_states, multi_tables,multi_years,multi_partial)])
def test_get_agencies_name_match(state, table_type, year, partial):
	if check_for_dataset(state, table_type):
		src = data.Source(state)
		try:
			agencies = src.get_agencies(partial_name=partial, table_type=table_type, year=year)
		except OPD_MinVersionError:
			return
		except Exception as e:
			raise

		assert len(agencies) > 1
			
			
def test_agency_filter():
	if check_for_dataset('New York', 'TRAFFIC CITATIONS'):
		src = data.Source("New York")
		agency="BUFFALO POLICE DEPT"
		# For speed, set private limit parameter so that only a single entry is requested
		table = src.load('TRAFFIC CITATIONS', 2021, agency=agency, pbar=False, nrows=100)
		
		assert len(table.table)==100
		assert table.table[table.agency_field].nunique()==1
		assert table.table.iloc[0][table.agency_field] == agency


def test_to_csv():
	if check_for_dataset('New York', 'TRAFFIC CITATIONS'):
		src = data.Source("New York")
		agency="BUFFALO POLICE DEPT"
		year = 2021
		table = src.load('TRAFFIC CITATIONS', 2021, agency=agency, pbar=False, nrows=100)

		table.to_csv()

		filename = table.get_csv_filename()
		assert os.path.exists(filename)

		try:
			# Load table back in
			src.load_from_csv(year, agency=agency)
		except:
			raise
		finally:
			os.remove(filename)


def is_filterable(data_type):
	data_type = DataType(data_type)
	if data_type in [DataType.CSV, DataType.EXCEL]:
		return False
	elif data_type in [DataType.ArcGIS, DataType.SOCRATA, DataType.CARTO, DataType.CKAN]:
		return True
	else:
		raise ValueError("Unknown table type")

def is_stanford(url):
	return "stanford.edu" in url

def log_errors_to_file(*args):
	if not os.path.exists(log_folder):
		os.mkdir(log_folder)

	filename = os.path.join(log_folder, log_filename)

	if os.path.exists(filename):
		perm = "r+"
	else:
		perm = "w"

	with open(filename, perm) as f:
		for x in args:
			for e in x:
				new_line = ', '.join([str(x) for x in e.args])
				skip = False
				if perm == "r+":
					for line in f:
						if new_line in line or line in new_line:
							skip = True
							break

				if not skip:
					f.write(new_line)
					f.write("\n")

if __name__ == "__main__":
	from test_utils import get_datasets
	# For testing
	# (csvfile, source, last, skip, loghtml)

	use_changed_rows = False

	csvfile = None
	# csvfile = os.path.join("..","opd-data","opd_source_table.csv")
	start_idx = 0
	source = None
	# source = "Burlington"
	skip = None
	# skip = "Sacramento,Beloit,Rutland"
	
	datasets = get_datasets(csvfile, use_changed_rows)

	# skip = "Corona"
	test_get_agencies(datasets, source, start_idx, skip)
	# test_agency_filter(datasets, None, None, skip, None)
	# test_to_csv(datasets, None, None, skip, None)
	test_get_years(datasets, source, start_idx, skip, None)



import warnings
from pandas import DataFrame, Series
from pandas.core.indexing import _iLocIndexer, _LocIndexer
from .messages import CIV_DEPRECATION_MESSAGE, CIV_REPLACEMENT_MESSAGE

def _cast_if_necessary(result):
    if isinstance(result, Series) and result.name == 'TableType' and \
        result.str.contains("SUBJECT").any():
        result = DeprecationHandlerSeries(result)
    elif isinstance(result, DataFrame) and 'TableType' in result and \
        result["TableType"].str.contains("SUBJECT").any():
        result = DeprecationHandlerDataFrame(result)

    return result

class _iLocIndexerSub(_iLocIndexer):
    def __getitem__(self, key):
        result = super().__getitem__(key)

        if isinstance(key, tuple) and len(key)>1:
            try:
                # 8 is where the Year column has been moved to and therefore, any old iloc calls
                # requesting data past index 8 could now return the wrong value
                if (isinstance(key[1], int) and 8<=key[1]<13) or \
                    (isinstance(key[1], slice) and any([8<=x<13 for x in range(*key[1].indices(1000000))])) or \
                    (not isinstance(key[1], (int, slice)) and any(8<=x<13 for x in key[1])):
                    warnings.warn(DeprecationWarning("The Year column of the datasets table has recently been moved to column 8 from column 12. "+\
                                                     f"Column input {key[1]} could be outputting a different column than desired for index 8-12 as a result "+\
                                                     "if the code was written prior to this change. Code may need to be changed. It is recommended "+\
                                                     "that .loc be used to index the datasets table rather than .iloc."))
            except:
                raise

        return _cast_if_necessary(result)
 

class _LocIndexerSub(_LocIndexer):
    def __getitem__(self, key):
        result = super().__getitem__(key)
        
        return _cast_if_necessary(result)
 

class DeprecationHandlerSeries(Series):
    def _cmp_method(self, other, op):
        if 'CIVILIAN' in other:
            new_other = other.replace("CIVILIAN","SUBJECT")
            warnings.warn(CIV_DEPRECATION_MESSAGE + CIV_REPLACEMENT_MESSAGE.format(other, new_other),
                            DeprecationWarning,
                            stacklevel=1)
            other = new_other

        return super()._cmp_method(other, op)
 

    def isin(self, values) -> Series:
        if self.name == 'TableType' and any(["CIVILIAN" in x for x in values]):
            new_values = [x.replace("CIVILIAN","SUBJECT") for x in values]
            warnings.warn(DeprecationWarning(CIV_DEPRECATION_MESSAGE + CIV_REPLACEMENT_MESSAGE.format(values, new_values)))
            values = new_values

        return super().isin(values)       


class DeprecationHandlerDataFrame(DataFrame):
    def __getitem__(self, key):
        result = super().__getitem__(key)
        return _cast_if_necessary(result)
    

    def query(self, *args, **kwargs):
        result = super().query(*args, **kwargs)
        return _cast_if_necessary(result)
    

    def copy(self, deep = True):
        result = super().copy(deep)
        return DeprecationHandlerDataFrame(result)

    @property
    def iloc(self) -> _iLocIndexerSub:
        return _iLocIndexerSub("iloc", self)

    @property
    def loc(self) -> _LocIndexerSub:
        return _LocIndexerSub("loc", self)